# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a processor for reach and frequency optimization inference on a Meridian model.

This module provides the `ReachFrequencyOptimizationProcessor`, which optimizes
the average frequency for reach and frequency (R&F) media channels in a trained
Meridian model to maximize ROI.

The processor takes a trained model and a `ReachFrequencyOptimizationSpec`
object. The spec defines the constraints for the optimization, such as the
minimum and maximum average frequency to consider for each channel.

Key Features:

-   Optimizes average frequency for all R&F channels simultaneously.
-   Allows setting minimum and maximum frequency constraints.
-   Generates detailed results, including the optimal average frequency for
    each channel, the expected outcomes at this optimal frequency, and
    response curves showing KPI/Revenue as a function of spend.
-   Outputs results in a structured protobuf format
    (`ReachFrequencyOptimization`).

Key Classes:

-   `ReachFrequencyOptimizationSpec`: Dataclass to specify optimization
    parameters and constraints.
-   `ReachFrequencyOptimizationProcessor`: The main processor class to execute
    the R&F optimization.

Example Usage:

```python
from schema.processors import reach_frequency_optimization_processor
from schema.processors import common
from schema.processors import model_processor
import datetime

# Assuming 'mmm' is a trained Meridian model object with R&F channels
trained_model = model_processor.TrainedModel(mmm)

spec = reach_frequency_optimization_processor.ReachFrequencyOptimizationSpec(
    optimization_name="rf_optimize_q1",
    start_date=datetime.date(2023, 1, 1),
    end_date=datetime.date(2023, 4, 1),
    min_frequency=1.0,
    max_frequency=10.0,  # Optional, defaults to model's max frequency
    kpi_type=common.KpiType.REVENUE,
)

processor = (
    reach_frequency_optimization_processor.ReachFrequencyOptimizationProcessor(
        trained_model
    )
)
# result is a rf_pb.ReachFrequencyOptimization proto
result = processor.execute([spec])

print(f"R&F Optimization results for {spec.optimization_name}:")
# Access results from the proto, e.g.:
# result.results[0].optimized_channel_frequencies
# result.results[0].optimized_marketing_analysis
# result.results[0].frequency_outcome_grid
```

Note: You can provide the processor with multiple specs. This would result in
a `ReachFrequencyOptimization` output with multiple results therein.
"""

from collections.abc import Sequence
import dataclasses

from meridian import backend
from meridian import constants
from mmm.v1 import mmm_pb2 as pb
from mmm.v1.common import kpi_type_pb2 as kpi_type_pb
from mmm.v1.marketing.analysis import marketing_analysis_pb2 as analysis_pb
from mmm.v1.marketing.analysis import media_analysis_pb2 as media_analysis_pb
from mmm.v1.marketing.analysis import outcome_pb2 as outcome_pb
from mmm.v1.marketing.analysis import response_curve_pb2
from mmm.v1.marketing.optimization import constraints_pb2 as constraints_pb
from mmm.v1.marketing.optimization import reach_frequency_optimization_pb2 as rf_pb
from schema.processors import common
from schema.processors import model_processor
from schema.utils import time_record
import numpy as np
import xarray as xr


__all__ = [
    "ReachFrequencyOptimizationSpec",
    "ReachFrequencyOptimizationProcessor",
]


_STEP_SIZE_DECIMAL_PRECISION = 1
_STEP_SIZE = _STEP_SIZE_DECIMAL_PRECISION / 10
_TOL = 1e-6


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReachFrequencyOptimizationSpec(model_processor.OptimizationSpec):
  """Spec dataclass for marketing reach and frequency optimization processor.

  A frequency grid is generated using the range `[rounded_min_frequency,
  rounded_max_frequency]` and a step size of `STEP_SIZE=0.1`.
  `rounded_min_frequency` and `rounded_max_frequency` are rounded to the
  nearest multiple of `STEP_SIZE`.

  This spec is used both as user input to inform the R&F optimization processor
  of its constraints and parameters, as well as an output structure that is
  serializable to a `ReachFrequencyOptimizationSpec` proto. The latter serves as
  a metadata embedded in a `ReachFrequencyOptimizationResult`. The output spec
  in the proto reflects the actual numbers used to generate the reach and
  frequency optimization result.

  Attributes:
    min_frequency: The minimum frequency constraint for each channel. Must be
      greater than or equal to `1.0`. Defaults to `1.0`.
    max_frequency: The maximum frequency constraint for each channel. Must be
      greater than min_frequency. Defaults to None. If this value is set to
      None, the model's max frequency will be used.
    rf_channels: The R&F media channels in the model. When resolved with a
      model, the model's R&F channels will be present here. Ignored when used as
      input.
    kpi_type: A `common.KpiType` enum denoting whether the optimized KPI is of a
      `'revenue'` or `'non-revenue'` type.
  """

  min_frequency: float = 1.0
  max_frequency: float | None = None
  rf_channels: Sequence[str] = dataclasses.field(default_factory=list)
  kpi_type: common.KpiType = common.KpiType.REVENUE

  @property
  def selected_times(self) -> tuple[str | None, str | None] | None:
    """The start and end dates, as a tuple of date strings."""
    start, end = (None, None)
    if self.start_date is not None:
      start = self.start_date.strftime(constants.DATE_FORMAT)
    if self.end_date is not None:
      end = self.end_date.strftime(constants.DATE_FORMAT)

    if start or end:
      return (start, end)
    return None

  @property
  def objective(self) -> common.TargetMetric:
    """A Meridian budget optimization objective is always ROI."""
    return common.TargetMetric.ROI

  def validate(self):
    super().validate()
    if self.min_frequency < 0:
      raise ValueError("Min frequency must be non-negative.")
    if (
        self.max_frequency is not None
        and self.max_frequency < self.min_frequency
    ):
      raise ValueError("Max frequency must be greater than min frequency.")

  def to_proto(self) -> rf_pb.ReachFrequencyOptimizationSpec:
    # When invoked as an output proto, the spec should have been fully resolved.
    if self.start_date is None or self.end_date is None:
      raise ValueError(
          "Start and end dates must be resolved before this spec can be"
          " serialized."
      )

    return rf_pb.ReachFrequencyOptimizationSpec(
        date_interval=time_record.create_date_interval_pb(
            self.start_date, self.end_date, tag=self.date_interval_tag
        ),
        rf_channel_constraints=[
            rf_pb.RfChannelConstraint(
                channel_name=channel,
                frequency_constraint=constraints_pb.FrequencyConstraint(
                    min_frequency=self.min_frequency,
                    max_frequency=self.max_frequency,
                ),
            )
            for channel in self.rf_channels
        ],
        objective=self.objective.value,
        kpi_type=(
            kpi_type_pb.KpiType.REVENUE
            if self.kpi_type == common.KpiType.REVENUE
            else kpi_type_pb.KpiType.NON_REVENUE
        ),
    )


class ReachFrequencyOptimizationProcessor(
    model_processor.ModelProcessor[
        ReachFrequencyOptimizationSpec, rf_pb.ReachFrequencyOptimization
    ],
):
  """A Processor for marketing reach and frequency optimization."""

  def __init__(
      self,
      trained_model: model_processor.ModelType,
  ):
    trained_model = model_processor.ensure_trained_model(trained_model)
    self._internal_analyzer = trained_model.internal_analyzer
    self._meridian = trained_model.mmm

    if trained_model.mmm.input_data.rf_channel is None:
      raise ValueError("RF channels must be set in the model.")

    self._all_rf_channels = trained_model.mmm.input_data.rf_channel.data

  @classmethod
  def spec_type(cls) -> type[ReachFrequencyOptimizationSpec]:
    return ReachFrequencyOptimizationSpec

  @classmethod
  def output_type(cls) -> type[rf_pb.ReachFrequencyOptimization]:
    return rf_pb.ReachFrequencyOptimization

  def _to_target_precision(self, value: float) -> float:
    return round(value, _STEP_SIZE_DECIMAL_PRECISION)

  def _set_output(
      self, output: pb.Mmm, result: rf_pb.ReachFrequencyOptimization
  ):
    output.marketing_optimization.reach_frequency_optimization.CopyFrom(result)

  def execute(
      self, specs: Sequence[ReachFrequencyOptimizationSpec]
  ) -> rf_pb.ReachFrequencyOptimization:
    output = rf_pb.ReachFrequencyOptimization()

    group_ids = [spec.group_id for spec in specs if spec.group_id]
    if len(set(group_ids)) != len(group_ids):
      raise ValueError(
          "Specified group_id must be unique among the given group of specs."
      )

    for spec in specs:
      selected_times = spec.resolver(
          self._meridian.input_data.time_coordinates
      ).resolve_to_enumerated_selected_times()

      grid_min_freq = self._to_target_precision(spec.min_frequency)
      # If the max frequency is not set, use the model's max frequency.
      grid_max_freq = self._to_target_precision(
          spec.max_frequency or np.max(self._meridian.rf_tensors.frequency)
      )
      grid = [
          self._to_target_precision(f)
          for f in np.arange(grid_min_freq, grid_max_freq + _TOL, _STEP_SIZE)
      ]

      # Note that the internal analyzer, like the budget optimizer, maximizes
      # non-revenue KPI if input data is of non-revenue and the user selects
      # `use_kpi=True`. Otherwise, it maximizes revenue KPI.
      optimal_frequency = self._internal_analyzer.optimal_freq(
          selected_times=selected_times,
          confidence_level=spec.confidence_level,
          freq_grid=grid,
          use_kpi=(spec.kpi_type == common.KpiType.NON_REVENUE),
          max_frequency=spec.max_frequency,
      )
      response_curve = self._internal_analyzer.response_curves(
          confidence_level=spec.confidence_level,
          selected_times=selected_times,
          by_reach=False,
          use_kpi=(spec.kpi_type == common.KpiType.NON_REVENUE),
          use_optimal_frequency=True,
      )

      spend_data = self._compute_spend_data(selected_times=selected_times)

      # Obtain the output spec.
      start, end = spec.resolver(
          self._meridian.input_data.time_coordinates
      ).resolve_to_date_interval_open_end()

      # Copy the current spec, and resolve its date interval as well as model-
      # dependent parameters.
      output_spec = dataclasses.replace(
          spec,
          rf_channels=self._all_rf_channels,
          min_frequency=grid_min_freq,
          max_frequency=grid_max_freq,
          start_date=start,
          end_date=end,
      )

      output.results.append(
          self._to_reach_frequency_optimization_result(
              output_spec,
              optimal_frequency,
              response_curve,
              spend_data,
          )
      )
    return output

  def _compute_spend_data(
      self, selected_times: list[str] | None = None
  ) -> xr.Dataset:
    aggregated_spends = self._internal_analyzer.get_historical_spend(
        selected_times
    )
    aggregated_rf_spend = aggregated_spends.sel(
        {constants.CHANNEL: self._all_rf_channels}
    ).data
    total_spend = np.sum(aggregated_spends.data)
    pct_of_spend = 100.0 * aggregated_rf_spend / total_spend

    xr_dims = (constants.RF_CHANNEL,)
    xr_coords = {
        constants.RF_CHANNEL: (
            [constants.RF_CHANNEL],
            list(self._all_rf_channels),
        ),
    }
    xr_data_vars = {
        constants.SPEND: (xr_dims, aggregated_rf_spend),
        constants.PCT_OF_SPEND: (xr_dims, pct_of_spend),
    }

    return xr.Dataset(
        data_vars=xr_data_vars,
        coords=xr_coords,
    )

  def _to_reach_frequency_optimization_result(
      self,
      spec: ReachFrequencyOptimizationSpec,
      optimal_frequency: xr.Dataset,
      response_curve: xr.Dataset,
      spend_data: xr.Dataset,
  ) -> rf_pb.ReachFrequencyOptimizationResult:
    """Converts given optimal frequency dataset to protobuf form."""
    result = rf_pb.ReachFrequencyOptimizationResult(
        name=spec.optimization_name,
        spec=spec.to_proto(),
        optimized_channel_frequencies=_create_optimized_channel_frequencies(
            optimal_frequency
        ),
        optimized_marketing_analysis=self._to_marketing_analysis(
            spec,
            optimal_frequency,
            response_curve,
            spend_data,
        ),
        frequency_outcome_grid=self._create_frequency_outcome_grid(
            optimal_frequency,
            spec,
        ),
    )
    if spec.group_id:
      result.group_id = spec.group_id
    return result

  def _to_marketing_analysis(
      self,
      spec: ReachFrequencyOptimizationSpec,
      optimal_frequency: xr.Dataset,
      response_curve: xr.Dataset,
      spend_data: xr.Dataset,
  ) -> analysis_pb.MarketingAnalysis:
    """Converts an optimal frequency dataset to a `MarketingAnalysis` proto."""
    # `spec` should have been resolved with concrete date interval parameters.
    assert spec.start_date is not None and spec.end_date is not None

    optimized_marketing_analysis = analysis_pb.MarketingAnalysis(
        date_interval=time_record.create_date_interval_pb(
            start_date=spec.start_date,
            end_date=spec.end_date,
        ),
    )

    # Create a per-channel MediaAnalysis.
    channels = optimal_frequency.coords[constants.RF_CHANNEL].data
    for channel in channels:
      channel_optimal_frequency = optimal_frequency.sel(rf_channel=channel)
      channel_spend_data = spend_data.sel(rf_channel=channel)

      # TODO Add non-media analyses.
      channel_media_analysis = media_analysis_pb.MediaAnalysis(
          channel_name=channel,
          response_curve=_compute_response_curve(
              response_curve,
              channel,
          ),
          spend_info=media_analysis_pb.SpendInfo(
              spend=channel_spend_data[constants.SPEND].data.item(),
              spend_share=(
                  channel_spend_data[constants.PCT_OF_SPEND].data.item()
              ),
          ),
      )

      # Output one outcome per channel: either revenue or non-revenue.
      channel_media_analysis.media_outcomes.append(
          _to_outcome(
              channel_optimal_frequency,
              is_revenue_kpi=optimal_frequency.attrs[constants.IS_REVENUE_KPI],
          )
      )

      optimized_marketing_analysis.media_analyses.append(channel_media_analysis)

    return optimized_marketing_analysis

  def _create_frequency_outcome_grid(
      self,
      optimal_frequency_dataset: xr.Dataset,
      spec: ReachFrequencyOptimizationSpec,
  ) -> rf_pb.FrequencyOutcomeGrid:
    """Creates a FrequencyOutcomeGrid proto."""
    channel_cells = []
    frequencies = optimal_frequency_dataset.coords[constants.FREQUENCY].data
    channels = optimal_frequency_dataset.coords[constants.RF_CHANNEL].data
    input_tensor_dims = "gtc"
    output_tensor_dims = "c"

    for channel in channels:
      cells = []
      for frequency in frequencies:
        new_frequency = (
            backend.ones_like(self._meridian.rf_tensors.frequency) * frequency
        )
        new_reach = (
            self._meridian.rf_tensors.frequency
            * self._meridian.rf_tensors.reach
            / new_frequency
        )
        channel_mask = [c == channel for c in channels]
        filtered_reach = backend.boolean_mask(new_reach, channel_mask, axis=2)
        aggregated_reach = backend.einsum(
            f"{input_tensor_dims}->...{output_tensor_dims}", filtered_reach
        )
        reach = aggregated_reach.numpy()[-1]

        metric_data_array = optimal_frequency_dataset[constants.ROI].sel(
            frequency=frequency, rf_channel=channel
        )
        outcome = common.to_estimate(metric_data_array, spec.confidence_level)

        cell = rf_pb.FrequencyOutcomeGrid.Cell(
            outcome=outcome,
            reach_frequency=rf_pb.ReachFrequency(
                reach=int(reach),
                average_frequency=frequency,
            ),
        )
        cells.append(cell)

      channel_cell = rf_pb.FrequencyOutcomeGrid.ChannelCells(
          channel_name=channel,
          cells=cells,
      )
      channel_cells.append(channel_cell)

    return rf_pb.FrequencyOutcomeGrid(
        name=spec.grid_name,
        frequency_step_size=_STEP_SIZE,
        channel_cells=channel_cells,
    )


def _create_optimized_channel_frequencies(
    optimal_frequency_dataset: xr.Dataset,
) -> list[rf_pb.OptimizedChannelFrequency]:
  """Creates an OptimizedChannelFrequency proto for each channel in the dataset."""
  optimal_frequency_protos = []
  optimal_frequency = optimal_frequency_dataset[constants.OPTIMAL_FREQUENCY]
  channels = optimal_frequency.coords[constants.RF_CHANNEL].data

  for channel in channels:
    optimal_frequency_protos.append(
        rf_pb.OptimizedChannelFrequency(
            channel_name=channel,
            optimal_average_frequency=optimal_frequency.sel(
                rf_channel=channel
            ).item(),
        )
    )
  return optimal_frequency_protos


def _to_outcome(
    channel_optimal_frequency: xr.Dataset,
    is_revenue_kpi: bool,
) -> outcome_pb.Outcome:
  """Returns an `Outcome` value for a given channel's optimized media analysis.

  Args:
    channel_optimal_frequency: A channel-selected dataset from
      `Analyzer.optimal_freq()`.
    is_revenue_kpi: Whether the KPI is revenue-based.
  """
  confidence_level = channel_optimal_frequency.attrs[constants.CONFIDENCE_LEVEL]
  return outcome_pb.Outcome(
      kpi_type=(
          kpi_type_pb.REVENUE if is_revenue_kpi else kpi_type_pb.NON_REVENUE
      ),
      roi=common.to_estimate(
          channel_optimal_frequency.optimized_roi, confidence_level
      ),
      marginal_roi=common.to_estimate(
          channel_optimal_frequency.optimized_mroi_by_frequency,
          confidence_level,
      ),
      cost_per_contribution=common.to_estimate(
          channel_optimal_frequency.optimized_cpik,
          confidence_level=confidence_level,
      ),
      contribution=outcome_pb.Contribution(
          value=common.to_estimate(
              channel_optimal_frequency.optimized_incremental_outcome,
              confidence_level,
          ),
      ),
      effectiveness=outcome_pb.Effectiveness(
          media_unit=constants.IMPRESSIONS,
          value=common.to_estimate(
              channel_optimal_frequency.optimized_effectiveness,
              confidence_level,
          ),
      ),
  )


def _compute_response_curve(
    response_curve_dataset: xr.Dataset,
    channel_name: str,
) -> response_curve_pb2.ResponseCurve:
  """Returns a ResponseCurve proto for the given channel.

  Args:
    response_curve_dataset: A dataset containing the data needed to generate a
      response curve.
    channel_name: The name of the channel to analyze.
  """

  spend_multiplier_list = response_curve_dataset.coords[
      constants.SPEND_MULTIPLIER
  ].data
  response_points: list[response_curve_pb2.ResponsePoint] = []

  for spend_multiplier in spend_multiplier_list:
    spend = (
        response_curve_dataset[constants.SPEND]
        .sel(spend_multiplier=spend_multiplier, channel=channel_name)
        .data.item()
    )
    incremental_outcome = (
        response_curve_dataset[constants.INCREMENTAL_OUTCOME]
        .sel(
            spend_multiplier=spend_multiplier,
            channel=channel_name,
            metric=constants.MEAN,
        )
        .data.item()
    )

    response_point = response_curve_pb2.ResponsePoint(
        input_value=spend,
        incremental_kpi=incremental_outcome,
    )
    response_points.append(response_point)

  return response_curve_pb2.ResponseCurve(
      input_name=constants.SPEND,
      response_points=response_points,
  )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates an `Mmm` (Marketing Mix Model) proto for Meridian.

The MMM proto schema contains parts collected from the core model as well as
analysis results from trained model processors.
"""

from collections.abc import Sequence
from typing import TypeVar

from meridian.model import model
from mmm.v1 import mmm_pb2 as mmm_pb
from schema import model_consumer
from schema.processors import budget_optimization_processor
from schema.processors import marketing_processor
from schema.processors import model_fit_processor
from schema.processors import model_processor
from schema.processors import reach_frequency_optimization_processor


__all__ = [
    "create_mmm_proto",
]


_TYPES = (
    model_fit_processor.ModelFitProcessor,
    marketing_processor.MarketingProcessor,
    budget_optimization_processor.BudgetOptimizationProcessor,
    reach_frequency_optimization_processor.ReachFrequencyOptimizationProcessor,
)

SpecType = TypeVar("SpecType", bound=model_processor.Spec)
DatedSpecType = TypeVar("DatedSpecType", bound=model_processor.DatedSpec)
OptimizationSpecType = TypeVar(
    "OptimizationSpecType", bound=model_processor.OptimizationSpec
)


def create_mmm_proto(
    mmm: model.Meridian,
    specs: Sequence[SpecType],
    model_id: str = "",
) -> mmm_pb.Mmm:
  """Creates a model schema and analyses for various time buckets.

  Args:
    mmm: A trained Meridian model. A trained model has its posterior
      distributions already sampled.
    specs: A sequence of specs that specify the analyses to run on the model.
    model_id: An optional model identifier.

  Returns:
    A proto containing the model kernel at rest and its analysis results given
    user specs.
  """
  consumer = model_consumer.ModelConsumer(_TYPES)
  return consumer(mmm, specs, model_id)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meridian EDA Engine."""

from __future__ import annotations

import dataclasses
import functools
import typing
from typing import Optional, Protocol, Sequence

from meridian import backend
from meridian import constants
from meridian.model import transformers
from meridian.model.eda import constants as eda_constants
from meridian.model.eda import eda_outcome
from meridian.model.eda import eda_spec
import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.stats import outliers_influence
import xarray as xr


if typing.TYPE_CHECKING:
  from meridian.model import model  # pylint: disable=g-bad-import-order,g-import-not-at-top

__all__ = ['EDAEngine', 'GeoLevelCheckOnNationalModelError']

_DEFAULT_DA_VAR_AGG_FUNCTION = np.sum
_CORRELATION_COL_NAME = eda_constants.CORRELATION
_STACK_VAR_COORD_NAME = eda_constants.VARIABLE
_CORR_VAR1 = eda_constants.VARIABLE_1
_CORR_VAR2 = eda_constants.VARIABLE_2
_CORRELATION_MATRIX_NAME = 'correlation_matrix'
_OVERALL_PAIRWISE_CORR_THRESHOLD = 0.999
_GEO_PAIRWISE_CORR_THRESHOLD = 0.999
_NATIONAL_PAIRWISE_CORR_THRESHOLD = 0.999
_Q1_THRESHOLD = 0.25
_Q3_THRESHOLD = 0.75
_IQR_MULTIPLIER = 1.5
_STD_WITH_OUTLIERS_VAR_NAME = 'std_with_outliers'
_STD_WITHOUT_OUTLIERS_VAR_NAME = 'std_without_outliers'
_STD_THRESHOLD = 1e-4
_OUTLIERS_COL_NAME = 'outliers'
_ABS_OUTLIERS_COL_NAME = 'abs_outliers'
_VIF_COL_NAME = 'VIF'


class _NamedEDACheckCallable(Protocol):
  """A callable that returns an EDAOutcome and has a __name__ attribute."""

  __name__: str

  def __call__(self) -> eda_outcome.EDAOutcome:
    ...


class GeoLevelCheckOnNationalModelError(Exception):
  """Raised when a geo-level check is called on a national model."""

  pass


@dataclasses.dataclass(frozen=True)
class _RFNames:
  """Holds constant names for reach and frequency data arrays."""

  reach: str
  reach_scaled: str
  frequency: str
  impressions: str
  impressions_scaled: str
  national_reach: str
  national_reach_scaled: str
  national_frequency: str
  national_impressions: str
  national_impressions_scaled: str


_ORGANIC_RF_NAMES = _RFNames(
    reach=constants.ORGANIC_REACH,
    reach_scaled=constants.ORGANIC_REACH_SCALED,
    frequency=constants.ORGANIC_FREQUENCY,
    impressions=constants.ORGANIC_RF_IMPRESSIONS,
    impressions_scaled=constants.ORGANIC_RF_IMPRESSIONS_SCALED,
    national_reach=constants.NATIONAL_ORGANIC_REACH,
    national_reach_scaled=constants.NATIONAL_ORGANIC_REACH_SCALED,
    national_frequency=constants.NATIONAL_ORGANIC_FREQUENCY,
    national_impressions=constants.NATIONAL_ORGANIC_RF_IMPRESSIONS,
    national_impressions_scaled=constants.NATIONAL_ORGANIC_RF_IMPRESSIONS_SCALED,
)


_RF_NAMES = _RFNames(
    reach=constants.REACH,
    reach_scaled=constants.REACH_SCALED,
    frequency=constants.FREQUENCY,
    impressions=constants.RF_IMPRESSIONS,
    impressions_scaled=constants.RF_IMPRESSIONS_SCALED,
    national_reach=constants.NATIONAL_REACH,
    national_reach_scaled=constants.NATIONAL_REACH_SCALED,
    national_frequency=constants.NATIONAL_FREQUENCY,
    national_impressions=constants.NATIONAL_RF_IMPRESSIONS,
    national_impressions_scaled=constants.NATIONAL_RF_IMPRESSIONS_SCALED,
)


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReachFrequencyData:
  """Holds reach and frequency data arrays.

  Attributes:
    reach_raw_da: Raw reach data.
    reach_scaled_da: Scaled reach data.
    national_reach_raw_da: National raw reach data.
    national_reach_scaled_da: National scaled reach data.
    frequency_da: Frequency data.
    national_frequency_da: National frequency data.
    rf_impressions_scaled_da: Scaled reach * frequency impressions data.
    national_rf_impressions_scaled_da: National scaled reach * frequency
      impressions data.
    rf_impressions_raw_da: Raw reach * frequency impressions data.
    national_rf_impressions_raw_da: National raw reach * frequency impressions
      data.
  """

  reach_raw_da: xr.DataArray
  reach_scaled_da: xr.DataArray
  national_reach_raw_da: xr.DataArray
  national_reach_scaled_da: xr.DataArray
  frequency_da: xr.DataArray
  national_frequency_da: xr.DataArray
  rf_impressions_scaled_da: xr.DataArray
  national_rf_impressions_scaled_da: xr.DataArray
  rf_impressions_raw_da: xr.DataArray
  national_rf_impressions_raw_da: xr.DataArray


def _data_array_like(
    *, da: xr.DataArray, values: np.ndarray | backend.Tensor
) -> xr.DataArray:
  """Returns a DataArray from `values` with the same structure as `da`.

  Args:
    da: The DataArray whose structure (dimensions, coordinates, name, and attrs)
      will be used for the new DataArray.
    values: The numpy array or backend tensor to use as the values for the new
      DataArray.

  Returns:
    A new DataArray with the provided `values` and the same structure as `da`.
  """
  return xr.DataArray(
      values,
      coords=da.coords,
      dims=da.dims,
      name=da.name,
      attrs=da.attrs,
  )


def stack_variables(
    ds: xr.Dataset, coord_name: str = _STACK_VAR_COORD_NAME
) -> xr.DataArray:
  """Stacks data variables of a Dataset into a single DataArray.

  This function is designed to work with Datasets that have 'time' or 'geo'
  dimensions, which are preserved. Other dimensions are stacked into a new
  dimension.

  Args:
    ds: The input xarray.Dataset to stack.
    coord_name: The name of the new coordinate for the stacked dimension.

  Returns:
    An xarray.DataArray with the specified dimensions stacked.
  """
  dims = []
  coords = []
  sample_dims = []
  # Dimensions have the same names as the coordinates.
  for dim in ds.dims:
    if dim in [constants.TIME, constants.GEO]:
      sample_dims.append(dim)
      continue
    dims.append(dim)
    coords.extend(ds.coords[dim].values.tolist())

  da = ds.to_stacked_array(coord_name, sample_dims=sample_dims)
  da = da.reset_index(dims, drop=True).assign_coords({coord_name: coords})
  return da


def _compute_correlation_matrix(
    input_da: xr.DataArray, dims: str | Sequence[str]
) -> xr.DataArray:
  """Computes the correlation matrix for variables in a DataArray.

  Args:
    input_da: An xr.DataArray containing variables for which to compute
      correlations.
    dims: Dimensions along which to compute correlations. Can only be TIME or
      GEO.

  Returns:
    An xr.DataArray containing the correlation matrix.
  """
  # Create two versions for correlation
  da1 = input_da.rename({_STACK_VAR_COORD_NAME: _CORR_VAR1})
  da2 = input_da.rename({_STACK_VAR_COORD_NAME: _CORR_VAR2})

  # Compute pairwise correlation across dims. Other dims are broadcasted.
  corr_mat_da = xr.corr(da1, da2, dim=dims)
  corr_mat_da.name = _CORRELATION_MATRIX_NAME
  return corr_mat_da


def _get_upper_triangle_corr_mat(corr_mat_da: xr.DataArray) -> xr.DataArray:
  """Gets the upper triangle of a correlation matrix.

  Args:
    corr_mat_da: An xr.DataArray containing the correlation matrix.

  Returns:
    An xr.DataArray containing only the elements in the upper triangle of the
    correlation matrix, with other elements masked as NaN.
  """
  n_vars = corr_mat_da.sizes[_CORR_VAR1]
  mask_np = np.triu(np.ones((n_vars, n_vars), dtype=bool), k=1)
  mask = xr.DataArray(
      mask_np,
      dims=[_CORR_VAR1, _CORR_VAR2],
      coords={
          _CORR_VAR1: corr_mat_da[_CORR_VAR1],
          _CORR_VAR2: corr_mat_da[_CORR_VAR2],
      },
  )
  return corr_mat_da.where(mask)


def _find_extreme_corr_pairs(
    extreme_corr_da: xr.DataArray, extreme_corr_threshold: float
) -> pd.DataFrame:
  """Finds extreme correlation pairs in a correlation matrix."""
  corr_tri = _get_upper_triangle_corr_mat(extreme_corr_da)
  extreme_corr_da = corr_tri.where(abs(corr_tri) > extreme_corr_threshold)

  return (
      extreme_corr_da.to_dataframe(name=_CORRELATION_COL_NAME)
      .dropna()
      .assign(**{
          eda_constants.ABS_CORRELATION_COL_NAME: (
              lambda x: x[_CORRELATION_COL_NAME].abs()
          )
      })
      .sort_values(
          by=eda_constants.ABS_CORRELATION_COL_NAME,
          ascending=False,
          inplace=False,
      )
  )


def _get_outlier_bounds(
    input_da: xr.DataArray,
) -> tuple[xr.DataArray, xr.DataArray]:
  """Computes lower and upper bounds for outliers across time using the IQR method.

  Args:
    input_da: A DataArray for which to calculate outlier bounds.

  Returns:
    A tuple containing the lower and upper bounds of outliers as DataArrays.
  """
  # TODO: Allow users to specify custom outlier definitions.
  q1 = input_da.quantile(_Q1_THRESHOLD, dim=constants.TIME)
  q3 = input_da.quantile(_Q3_THRESHOLD, dim=constants.TIME)
  iqr = q3 - q1
  lower_bound = q1 - _IQR_MULTIPLIER * iqr
  upper_bound = q3 + _IQR_MULTIPLIER * iqr
  return lower_bound, upper_bound


def _calculate_std(
    input_da: xr.DataArray,
) -> xr.Dataset:
  """Helper function to compute std with and without outliers.

  Args:
    input_da: A DataArray for which to calculate the std.

  Returns:
    A Dataset with two data variables: 'std_with_outliers' and
    'std_without_outliers'.
  """
  std_with_outliers = input_da.std(dim=constants.TIME, ddof=1)

  lower_bound, upper_bound = _get_outlier_bounds(input_da)
  da_no_outlier = input_da.where(
      (input_da >= lower_bound) & (input_da <= upper_bound)
  )
  std_without_outliers = da_no_outlier.std(dim=constants.TIME, ddof=1)

  std_ds = xr.Dataset({
      _STD_WITH_OUTLIERS_VAR_NAME: std_with_outliers,
      _STD_WITHOUT_OUTLIERS_VAR_NAME: std_without_outliers,
  })
  return std_ds


def _calculate_outliers(
    input_da: xr.DataArray,
) -> pd.DataFrame:
  """Helper function to extract outliers from a DataArray across time.

  Args:
    input_da: A DataArray from which to extract outliers.

  Returns:
    A DataFrame with columns for variables, geo (if applicable), time, and
    outlier values.
  """
  lower_bound, upper_bound = _get_outlier_bounds(input_da)
  outlier_da = input_da.where(
      (input_da < lower_bound) | (input_da > upper_bound)
  )
  outlier_df = (
      outlier_da.to_dataframe(name=_OUTLIERS_COL_NAME)
      .dropna()
      .assign(
          **{_ABS_OUTLIERS_COL_NAME: lambda x: np.abs(x[_OUTLIERS_COL_NAME])}
      )
      .sort_values(by=_ABS_OUTLIERS_COL_NAME, ascending=False, inplace=False)
  )
  return outlier_df


def _calculate_vif(input_da: xr.DataArray, var_dim: str) -> xr.DataArray:
  """Helper function to compute variance inflation factor.

  Args:
    input_da: A DataArray for which to calculate the VIF over sample dimensions
      (e.g. time and geo if applicable).
    var_dim: The dimension name of the variable to compute VIF for.

  Returns:
    A DataArray containing the VIF for each variable in the variable dimension.
  """
  num_vars = input_da.sizes[var_dim]
  np_data = input_da.values.reshape(-1, num_vars)
  np_data_with_const = sm.add_constant(
      np_data, prepend=True, has_constant='add'
  )

  # Compute VIF for each variable excluding const which is the first one in the
  # 'variable' dimension.
  vifs = [
      outliers_influence.variance_inflation_factor(np_data_with_const, i)
      for i in range(1, num_vars + 1)
  ]

  vif_da = xr.DataArray(
      vifs,
      coords={var_dim: input_da[var_dim].values},
      dims=[var_dim],
  )
  return vif_da


def _check_cost_media_unit_inconsistency(
    cost_da: xr.DataArray,
    media_units_da: xr.DataArray,
) -> pd.DataFrame:
  """Checks for inconsistencies between cost and media units.

  Args:
    cost_da: DataArray containing cost data.
    media_units_da: DataArray containing media unit data.

  Returns:
    A DataFrame of inconsistencies where either cost is zero and media units
    are
    positive, or cost is positive and media units are zero.
  """
  cost_media_units_ds = xr.merge([cost_da, media_units_da])

  # Condition 1: cost == 0 and media unit > 0
  zero_cost_positive_mask = (cost_da == 0) & (media_units_da > 0)
  zero_cost_positive_media_unit_df = (
      cost_media_units_ds.where(zero_cost_positive_mask).to_dataframe().dropna()
  )

  # Condition 2: cost > 0 and media unit == 0
  positive_cost_zero_mask = (cost_da > 0) & (media_units_da == 0)
  positive_cost_zero_media_unit_df = (
      cost_media_units_ds.where(positive_cost_zero_mask).to_dataframe().dropna()
  )

  return pd.concat(
      [zero_cost_positive_media_unit_df, positive_cost_zero_media_unit_df]
  )


def _check_cost_per_media_unit(
    cost_ds: xr.Dataset,
    media_units_ds: xr.Dataset,
    level: eda_outcome.AnalysisLevel,
) -> eda_outcome.EDAOutcome[eda_outcome.CostPerMediaUnitArtifact]:
  """Helper to check if the cost per media unit is valid."""
  findings = []
  # Stack variables with the same dimension name, so that they can be operated
  # on together.
  cost_da = stack_variables(cost_ds, constants.CHANNEL).rename(constants.SPEND)
  media_units_da = stack_variables(media_units_ds, constants.CHANNEL).rename(
      constants.MEDIA_UNITS
  )

  cost_media_unit_inconsistency_df = _check_cost_media_unit_inconsistency(
      cost_da,
      media_units_da,
  )
  if not cost_media_unit_inconsistency_df.empty:
    findings.append(
        eda_outcome.EDAFinding(
            severity=eda_outcome.EDASeverity.ATTENTION,
            explanation=(
                'There are instances of inconsistent cost and media units.'
                ' This occurs when cost is zero but media units are positive,'
                ' or when cost is positive but media units are zero. Please'
                ' review the outcome artifact for more details.'
            ),
        )
    )

  # Calculate cost per media unit
  # Avoid division by zero by setting cost to NaN where media units are 0.
  # Note that both (cost == media unit == 0) and (cost > 0 and media unit ==
  # 0) result in NaN, while the latter one is not desired.
  cost_per_media_unit_da = xr.where(
      media_units_da == 0,
      np.nan,
      cost_da / media_units_da,
  )
  cost_per_media_unit_da.name = eda_constants.COST_PER_MEDIA_UNIT

  outlier_df = _calculate_outliers(cost_per_media_unit_da)
  if not outlier_df.empty:
    findings.append(
        eda_outcome.EDAFinding(
            severity=eda_outcome.EDASeverity.ATTENTION,
            explanation=(
                'There are outliers in cost per media unit across time.'
                ' Please review the outcome artifact for more details.'
            ),
        )
    )

  # If no specific findings, add an INFO finding.
  if not findings:
    findings.append(
        eda_outcome.EDAFinding(
            severity=eda_outcome.EDASeverity.INFO,
            explanation='Please review the cost per media unit data.',
        )
    )

  artifact = eda_outcome.CostPerMediaUnitArtifact(
      level=level,
      cost_per_media_unit_da=cost_per_media_unit_da,
      cost_media_unit_inconsistency_df=cost_media_unit_inconsistency_df,
      outlier_df=outlier_df,
  )

  return eda_outcome.EDAOutcome(
      check_type=eda_outcome.EDACheckType.COST_PER_MEDIA_UNIT,
      findings=findings,
      analysis_artifacts=[artifact],
  )


class EDAEngine:
  """Meridian EDA Engine."""

  def __init__(
      self,
      meridian: model.Meridian,
      spec: eda_spec.EDASpec = eda_spec.EDASpec(),
  ):
    self._meridian = meridian
    self._spec = spec
    self._agg_config = self._spec.aggregation_config

  @property
  def spec(self) -> eda_spec.EDASpec:
    return self._spec

  @property
  def _is_national_data(self) -> bool:
    return self._meridian.is_national

  @functools.cached_property
  def controls_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled controls data array."""
    if self._meridian.input_data.controls is None:
      return None
    controls_scaled_da = _data_array_like(
        da=self._meridian.input_data.controls,
        values=self._meridian.controls_scaled,
    )
    controls_scaled_da.name = constants.CONTROLS_SCALED
    return controls_scaled_da

  @functools.cached_property
  def national_controls_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled controls data array."""
    if self._meridian.input_data.controls is None:
      return None
    if self._is_national_data:
      if self.controls_scaled_da is None:
        # This case should be impossible given the check above.
        raise RuntimeError(
            'controls_scaled_da is None when controls is not None.'
        )
      national_da = self.controls_scaled_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_CONTROLS_SCALED
    else:
      national_da = self._aggregate_and_scale_geo_da(
          self._meridian.input_data.controls,
          constants.NATIONAL_CONTROLS_SCALED,
          transformers.CenteringAndScalingTransformer,
          constants.CONTROL_VARIABLE,
          self._agg_config.control_variables,
      )
    return national_da

  @functools.cached_property
  def media_raw_da(self) -> xr.DataArray | None:
    """Returns the raw media data array."""
    if self._meridian.input_data.media is None:
      return None
    raw_media_da = self._truncate_media_time(self._meridian.input_data.media)
    raw_media_da.name = constants.MEDIA
    return raw_media_da

  @functools.cached_property
  def media_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled media data array."""
    if self._meridian.input_data.media is None:
      return None
    media_scaled_da = _data_array_like(
        da=self._meridian.input_data.media,
        values=self._meridian.media_tensors.media_scaled,
    )
    media_scaled_da.name = constants.MEDIA_SCALED
    return self._truncate_media_time(media_scaled_da)

  @functools.cached_property
  def media_spend_da(self) -> xr.DataArray | None:
    """Returns media spend.

    If the input spend is aggregated, it is allocated across geo and time
    proportionally to media units.
    """
    # No need to truncate the media time for media spend.
    da = self._meridian.input_data.allocated_media_spend
    if da is None:
      return None
    da = da.copy()
    da.name = constants.MEDIA_SPEND
    return da

  @functools.cached_property
  def national_media_spend_da(self) -> xr.DataArray | None:
    """Returns the national media spend data array."""
    media_spend = self.media_spend_da
    if media_spend is None:
      return None
    if self._is_national_data:
      national_da = media_spend.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_MEDIA_SPEND
    else:
      national_da = self._aggregate_and_scale_geo_da(
          self._meridian.input_data.allocated_media_spend,
          constants.NATIONAL_MEDIA_SPEND,
          None,
      )
    return national_da

  @functools.cached_property
  def national_media_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw media data array."""
    if self.media_raw_da is None:
      return None
    if self._is_national_data:
      national_da = self.media_raw_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_MEDIA
    else:
      # Note that media is summable by assumption.
      national_da = self._aggregate_and_scale_geo_da(
          self.media_raw_da,
          constants.NATIONAL_MEDIA,
          None,
      )
    return national_da

  @functools.cached_property
  def national_media_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled media data array."""
    if self.media_scaled_da is None:
      return None
    if self._is_national_data:
      national_da = self.media_scaled_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_MEDIA_SCALED
    else:
      # Note that media is summable by assumption.
      national_da = self._aggregate_and_scale_geo_da(
          self.media_raw_da,
          constants.NATIONAL_MEDIA_SCALED,
          transformers.MediaTransformer,
      )
    return national_da

  @functools.cached_property
  def organic_media_raw_da(self) -> xr.DataArray | None:
    """Returns the raw organic media data array."""
    if self._meridian.input_data.organic_media is None:
      return None
    raw_organic_media_da = self._truncate_media_time(
        self._meridian.input_data.organic_media
    )
    raw_organic_media_da.name = constants.ORGANIC_MEDIA
    return raw_organic_media_da

  @functools.cached_property
  def organic_media_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled organic media data array."""
    if self._meridian.input_data.organic_media is None:
      return None
    organic_media_scaled_da = _data_array_like(
        da=self._meridian.input_data.organic_media,
        values=self._meridian.organic_media_tensors.organic_media_scaled,
    )
    organic_media_scaled_da.name = constants.ORGANIC_MEDIA_SCALED
    return self._truncate_media_time(organic_media_scaled_da)

  @functools.cached_property
  def national_organic_media_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw organic media data array."""
    if self.organic_media_raw_da is None:
      return None
    if self._is_national_data:
      national_da = self.organic_media_raw_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_ORGANIC_MEDIA
    else:
      # Note that organic media is summable by assumption.
      national_da = self._aggregate_and_scale_geo_da(
          self.organic_media_raw_da, constants.NATIONAL_ORGANIC_MEDIA, None
      )
    return national_da

  @functools.cached_property
  def national_organic_media_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled organic media data array."""
    if self.organic_media_scaled_da is None:
      return None
    if self._is_national_data:
      national_da = self.organic_media_scaled_da.squeeze(
          constants.GEO, drop=True
      )
      national_da.name = constants.NATIONAL_ORGANIC_MEDIA_SCALED
    else:
      # Note that organic media is summable by assumption.
      national_da = self._aggregate_and_scale_geo_da(
          self.organic_media_raw_da,
          constants.NATIONAL_ORGANIC_MEDIA_SCALED,
          transformers.MediaTransformer,
      )
    return national_da

  @functools.cached_property
  def non_media_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled non-media treatments data array."""
    if self._meridian.input_data.non_media_treatments is None:
      return None
    non_media_scaled_da = _data_array_like(
        da=self._meridian.input_data.non_media_treatments,
        values=self._meridian.non_media_treatments_normalized,
    )
    non_media_scaled_da.name = constants.NON_MEDIA_TREATMENTS_SCALED
    return non_media_scaled_da

  @functools.cached_property
  def national_non_media_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled non-media treatment data array."""
    if self._meridian.input_data.non_media_treatments is None:
      return None
    if self._is_national_data:
      if self.non_media_scaled_da is None:
        # This case should be impossible given the check above.
        raise RuntimeError(
            'non_media_scaled_da is None when non_media_treatments is not None.'
        )
      national_da = self.non_media_scaled_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_NON_MEDIA_TREATMENTS_SCALED
    else:
      national_da = self._aggregate_and_scale_geo_da(
          self._meridian.input_data.non_media_treatments,
          constants.NATIONAL_NON_MEDIA_TREATMENTS_SCALED,
          transformers.CenteringAndScalingTransformer,
          constants.NON_MEDIA_CHANNEL,
          self._agg_config.non_media_treatments,
      )
    return national_da

  @functools.cached_property
  def rf_spend_da(self) -> xr.DataArray | None:
    """Returns RF spend.

    If the input spend is aggregated, it is allocated across geo and time
    proportionally to RF impressions (reach * frequency).
    """
    da = self._meridian.input_data.allocated_rf_spend
    if da is None:
      return None
    da = da.copy()
    da.name = constants.RF_SPEND
    return da

  @functools.cached_property
  def national_rf_spend_da(self) -> xr.DataArray | None:
    """Returns the national RF spend data array."""
    rf_spend = self.rf_spend_da
    if rf_spend is None:
      return None
    if self._is_national_data:
      national_da = rf_spend.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_RF_SPEND
    else:
      national_da = self._aggregate_and_scale_geo_da(
          self._meridian.input_data.allocated_rf_spend,
          constants.NATIONAL_RF_SPEND,
          None,
      )
    return national_da

  @functools.cached_property
  def _rf_data(self) -> ReachFrequencyData | None:
    if self._meridian.input_data.reach is None:
      return None
    return self._get_rf_data(
        self._meridian.input_data.reach,
        self._meridian.input_data.frequency,
        is_organic=False,
    )

  @property
  def reach_raw_da(self) -> xr.DataArray | None:
    """Returns the raw reach data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.reach_raw_da

  @property
  def reach_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled reach data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.reach_scaled_da  # pytype: disable=attribute-error

  @property
  def national_reach_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw reach data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.national_reach_raw_da

  @property
  def national_reach_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled reach data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.national_reach_scaled_da  # pytype: disable=attribute-error

  @property
  def frequency_da(self) -> xr.DataArray | None:
    """Returns the frequency data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.frequency_da  # pytype: disable=attribute-error

  @property
  def national_frequency_da(self) -> xr.DataArray | None:
    """Returns the national frequency data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.national_frequency_da  # pytype: disable=attribute-error

  @property
  def rf_impressions_raw_da(self) -> xr.DataArray | None:
    """Returns the raw RF impressions data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.rf_impressions_raw_da  # pytype: disable=attribute-error

  @property
  def national_rf_impressions_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw RF impressions data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.national_rf_impressions_raw_da  # pytype: disable=attribute-error

  @property
  def rf_impressions_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled RF impressions data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.rf_impressions_scaled_da

  @property
  def national_rf_impressions_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled RF impressions data array."""
    if self._rf_data is None:
      return None
    return self._rf_data.national_rf_impressions_scaled_da

  @functools.cached_property
  def _organic_rf_data(self) -> ReachFrequencyData | None:
    if self._meridian.input_data.organic_reach is None:
      return None
    return self._get_rf_data(
        self._meridian.input_data.organic_reach,
        self._meridian.input_data.organic_frequency,
        is_organic=True,
    )

  @property
  def organic_reach_raw_da(self) -> xr.DataArray | None:
    """Returns the raw organic reach data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.reach_raw_da

  @property
  def organic_reach_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled organic reach data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.reach_scaled_da  # pytype: disable=attribute-error

  @property
  def national_organic_reach_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw organic reach data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.national_reach_raw_da

  @property
  def national_organic_reach_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled organic reach data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.national_reach_scaled_da  # pytype: disable=attribute-error

  @property
  def organic_rf_impressions_scaled_da(self) -> xr.DataArray | None:
    """Returns the scaled organic RF impressions data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.rf_impressions_scaled_da

  @property
  def national_organic_rf_impressions_scaled_da(self) -> xr.DataArray | None:
    """Returns the national scaled organic RF impressions data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.national_rf_impressions_scaled_da

  @property
  def organic_frequency_da(self) -> xr.DataArray | None:
    """Returns the organic frequency data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.frequency_da  # pytype: disable=attribute-error

  @property
  def national_organic_frequency_da(self) -> xr.DataArray | None:
    """Returns the national organic frequency data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.national_frequency_da  # pytype: disable=attribute-error

  @property
  def organic_rf_impressions_raw_da(self) -> xr.DataArray | None:
    """Returns the raw organic RF impressions data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.rf_impressions_raw_da

  @property
  def national_organic_rf_impressions_raw_da(self) -> xr.DataArray | None:
    """Returns the national raw organic RF impressions data array."""
    if self._organic_rf_data is None:
      return None
    return self._organic_rf_data.national_rf_impressions_raw_da

  @functools.cached_property
  def geo_population_da(self) -> xr.DataArray | None:
    """Returns the geo population data array."""
    if self._is_national_data:
      return None
    return xr.DataArray(
        self._meridian.population,
        coords={constants.GEO: self._meridian.input_data.geo.values},
        dims=[constants.GEO],
        name=constants.POPULATION,
    )

  @functools.cached_property
  def kpi_scaled_da(self) -> xr.DataArray:
    """Returns the scaled KPI data array."""
    scaled_kpi_da = _data_array_like(
        da=self._meridian.input_data.kpi,
        values=self._meridian.kpi_scaled,
    )
    scaled_kpi_da.name = constants.KPI_SCALED
    return scaled_kpi_da

  @functools.cached_property
  def _overall_scaled_kpi_invariability_artifact(
      self,
  ) -> eda_outcome.KpiInvariabilityArtifact:
    """Returns an artifact of overall scaled KPI invariability."""
    return eda_outcome.KpiInvariabilityArtifact(
        level=eda_outcome.AnalysisLevel.OVERALL,
        kpi_da=self.kpi_scaled_da,
        kpi_stdev=self.kpi_scaled_da.std(ddof=1),
    )

  @functools.cached_property
  def national_kpi_scaled_da(self) -> xr.DataArray:
    """Returns the national scaled KPI data array."""
    if self._is_national_data:
      national_da = self.kpi_scaled_da.squeeze(constants.GEO, drop=True)
      national_da.name = constants.NATIONAL_KPI_SCALED
    else:
      # Note that kpi is summable by assumption.
      national_da = self._aggregate_and_scale_geo_da(
          self._meridian.input_data.kpi,
          constants.NATIONAL_KPI_SCALED,
          transformers.CenteringAndScalingTransformer,
      )
    return national_da

  @functools.cached_property
  def treatment_control_scaled_ds(self) -> xr.Dataset:
    """Returns a Dataset containing all scaled treatments and controls.

    This includes media, RF impressions, organic media, organic RF impressions,
    non-media treatments, and control variables, all at the geo level.
    """
    to_merge = [
        da
        for da in [
            self.media_scaled_da,
            self.rf_impressions_scaled_da,
            self.organic_media_scaled_da,
            self.organic_rf_impressions_scaled_da,
            self.controls_scaled_da,
            self.non_media_scaled_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge, join='inner')

  @functools.cached_property
  def all_spend_ds(self) -> xr.Dataset:
    """Returns a Dataset containing all spend data.

    This includes media spend and rf spend.
    """
    to_merge = [
        da
        for da in [
            self.media_spend_da,
            self.rf_spend_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge, join='inner')

  @functools.cached_property
  def national_all_spend_ds(self) -> xr.Dataset:
    """Returns a Dataset containing all national spend data.

    This includes media spend and rf spend.
    """
    to_merge = [
        da
        for da in [
            self.national_media_spend_da,
            self.national_rf_spend_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge, join='inner')

  @functools.cached_property
  def _stacked_treatment_control_scaled_da(self) -> xr.DataArray:
    """Returns a stacked DataArray of treatment_control_scaled_ds."""
    da = stack_variables(self.treatment_control_scaled_ds)
    da.name = constants.TREATMENT_CONTROL_SCALED
    return da

  @functools.cached_property
  def national_treatment_control_scaled_ds(self) -> xr.Dataset:
    """Returns a Dataset containing all scaled treatments and controls.

    This includes media, RF impressions, organic media, organic RF impressions,
    non-media treatments, and control variables, all at the national level.
    """
    to_merge_national = [
        da
        for da in [
            self.national_media_scaled_da,
            self.national_rf_impressions_scaled_da,
            self.national_organic_media_scaled_da,
            self.national_organic_rf_impressions_scaled_da,
            self.national_controls_scaled_da,
            self.national_non_media_scaled_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge_national, join='inner')

  @functools.cached_property
  def _stacked_national_treatment_control_scaled_da(self) -> xr.DataArray:
    """Returns a stacked DataArray of national_treatment_control_scaled_ds."""
    da = stack_variables(self.national_treatment_control_scaled_ds)
    da.name = constants.NATIONAL_TREATMENT_CONTROL_SCALED
    return da

  @functools.cached_property
  def treatments_without_non_media_scaled_ds(self) -> xr.Dataset:
    """Returns a Dataset of scaled treatments excluding non-media."""
    return self.treatment_control_scaled_ds.drop_dims(
        [constants.NON_MEDIA_CHANNEL, constants.CONTROL_VARIABLE],
        errors='ignore',
    )

  @functools.cached_property
  def national_treatments_without_non_media_scaled_ds(self) -> xr.Dataset:
    """Returns a Dataset of national scaled treatments excluding non-media."""
    return self.national_treatment_control_scaled_ds.drop_dims(
        [constants.NON_MEDIA_CHANNEL, constants.CONTROL_VARIABLE],
        errors='ignore',
    )

  @functools.cached_property
  def all_reach_scaled_da(self) -> xr.DataArray | None:
    """Returns a DataArray containing all scaled reach data.

    This includes both paid and organic reach, concatenated along the RF_CHANNEL
    dimension.

    Returns:
      A DataArray containing all scaled reach data, or None if no RF or organic
      RF channels are present.
    """
    reach_das = []
    if self.reach_scaled_da is not None:
      reach_das.append(self.reach_scaled_da)
    if self.organic_reach_scaled_da is not None:
      reach_das.append(
          self.organic_reach_scaled_da.rename(
              {constants.ORGANIC_RF_CHANNEL: constants.RF_CHANNEL}
          )
      )
    if not reach_das:
      return None
    da = xr.concat(reach_das, dim=constants.RF_CHANNEL)
    da.name = constants.ALL_REACH_SCALED
    return da

  @functools.cached_property
  def all_freq_da(self) -> xr.DataArray | None:
    """Returns a DataArray containing all frequency data.

    This includes both paid and organic frequency, concatenated along the
    RF_CHANNEL dimension.

    Returns:
      A DataArray containing all frequency data, or None if no RF or organic
      RF channels are present.
    """
    freq_das = []
    if self.frequency_da is not None:
      freq_das.append(self.frequency_da)
    if self.organic_frequency_da is not None:
      freq_das.append(
          self.organic_frequency_da.rename(
              {constants.ORGANIC_RF_CHANNEL: constants.RF_CHANNEL}
          )
      )
    if not freq_das:
      return None
    da = xr.concat(freq_das, dim=constants.RF_CHANNEL)
    da.name = constants.ALL_FREQUENCY
    return da

  @functools.cached_property
  def national_all_reach_scaled_da(self) -> xr.DataArray | None:
    """Returns a DataArray containing all national-level scaled reach data.

    This includes both paid and organic reach, concatenated along the
    RF_CHANNEL dimension.

    Returns:
      A DataArray containing all national-level scaled reach data, or None if
      no RF or organic RF channels are present.
    """
    national_reach_das = []
    if self.national_reach_scaled_da is not None:
      national_reach_das.append(self.national_reach_scaled_da)
    national_organic_reach_scaled_da = self.national_organic_reach_scaled_da
    if national_organic_reach_scaled_da is not None:
      national_reach_das.append(
          national_organic_reach_scaled_da.rename(
              {constants.ORGANIC_RF_CHANNEL: constants.RF_CHANNEL}
          )
      )
    if not national_reach_das:
      return None
    da = xr.concat(national_reach_das, dim=constants.RF_CHANNEL)
    da.name = constants.NATIONAL_ALL_REACH_SCALED
    return da

  @functools.cached_property
  def national_all_freq_da(self) -> xr.DataArray | None:
    """Returns a DataArray containing all national-level frequency data.

    This includes both paid and organic frequency, concatenated along the
    RF_CHANNEL dimension.

    Returns:
      A DataArray containing all national-level frequency data, or None if no
      RF or organic RF channels are present.
    """
    national_freq_das = []
    if self.national_frequency_da is not None:
      national_freq_das.append(self.national_frequency_da)
    national_organic_frequency_da = self.national_organic_frequency_da
    if national_organic_frequency_da is not None:
      national_freq_das.append(
          national_organic_frequency_da.rename(
              {constants.ORGANIC_RF_CHANNEL: constants.RF_CHANNEL}
          )
      )
    if not national_freq_das:
      return None
    da = xr.concat(national_freq_das, dim=constants.RF_CHANNEL)
    da.name = constants.NATIONAL_ALL_FREQUENCY
    return da

  @functools.cached_property
  def paid_raw_media_units_ds(self) -> xr.Dataset:
    to_merge = [
        da
        for da in [
            self.media_raw_da,
            self.rf_impressions_raw_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge, join='inner')

  @functools.cached_property
  def national_paid_raw_media_units_ds(self) -> xr.Dataset:
    to_merge = [
        da
        for da in [
            self.national_media_raw_da,
            self.national_rf_impressions_raw_da,
        ]
        if da is not None
    ]
    return xr.merge(to_merge, join='inner')

  @property
  def _critical_checks(
      self,
  ) -> list[tuple[_NamedEDACheckCallable, eda_outcome.EDACheckType]]:
    """Returns a list of critical checks to be performed."""
    checks = [
        (
            self.check_overall_kpi_invariability,
            eda_outcome.EDACheckType.KPI_INVARIABILITY,
        ),
        (self.check_vif, eda_outcome.EDACheckType.MULTICOLLINEARITY),
        (
            self.check_pairwise_corr,
            eda_outcome.EDACheckType.PAIRWISE_CORRELATION,
        ),
    ]
    return checks

  def _truncate_media_time(self, da: xr.DataArray) -> xr.DataArray:
    """Truncates the first `start` elements of the media time of a variable."""
    # This should not happen. If it does, it means this function is mis-used.
    if constants.MEDIA_TIME not in da.coords:
      raise ValueError(
          f'Variable does not have a media time coordinate: {da.name}.'
      )

    start = self._meridian.n_media_times - self._meridian.n_times
    da = da.copy().isel({constants.MEDIA_TIME: slice(start, None)})
    da = da.rename({constants.MEDIA_TIME: constants.TIME})
    return da

  def _scale_xarray(
      self,
      xarray: xr.DataArray,
      transformer_class: Optional[type[transformers.TensorTransformer]],
      population: Optional[backend.Tensor] = None,
  ) -> xr.DataArray:
    """Scales xarray values with a TensorTransformer."""
    da = xarray.copy()

    if transformer_class is None:
      return da
    if population is None:
      population = backend.ones([1], dtype=backend.float32)
    if transformer_class is transformers.CenteringAndScalingTransformer:
      xarray_transformer = transformers.CenteringAndScalingTransformer(
          tensor=da.values, population=population
      )
    elif transformer_class is transformers.MediaTransformer:
      xarray_transformer = transformers.MediaTransformer(
          media=da.values, population=population
      )
    else:
      raise ValueError(
          'Unknown transformer class: '
          + str(transformer_class)
          + '.\nMust be one of: CenteringAndScalingTransformer or'
          ' MediaTransformer.'
      )
    da.values = xarray_transformer.forward(da.values)
    return da

  def _aggregate_variables(
      self,
      geo_da: xr.DataArray,
      channel_dim: str,
      da_var_agg_map: eda_spec.AggregationMap,
      keepdims: bool = True,
  ) -> xr.DataArray:
    """Aggregates variables within a DataArray based on user-defined functions.

    Args:
      geo_da: The geo-level DataArray containing multiple variables along
        channel_dim.
      channel_dim: The name of the dimension coordinate to aggregate over (e.g.,
        constants.CONTROL_VARIABLE).
      da_var_agg_map: A dictionary mapping dataArray variable names to
        aggregation functions.
      keepdims: Whether to keep the dimensions of the aggregated DataArray.

    Returns:
      An xr.DataArray aggregated to the national level, with each variable
      aggregated according to the da_var_agg_map.
    """
    agg_results = []
    for var_name in geo_da[channel_dim].values:
      var_data = geo_da.sel({channel_dim: var_name})
      agg_func = da_var_agg_map.get(var_name, _DEFAULT_DA_VAR_AGG_FUNCTION)
      # Apply the aggregation function over the GEO dimension
      aggregated_data = var_data.reduce(
          agg_func, dim=constants.GEO, keepdims=keepdims
      )
      agg_results.append(aggregated_data)

    # Combine the aggregated variables back into a single DataArray
    return xr.concat(agg_results, dim=channel_dim).transpose(..., channel_dim)

  def _aggregate_and_scale_geo_da(
      self,
      geo_da: xr.DataArray,
      national_da_name: str,
      transformer_class: Optional[type[transformers.TensorTransformer]],
      channel_dim: Optional[str] = None,
      da_var_agg_map: Optional[eda_spec.AggregationMap] = None,
  ) -> xr.DataArray:
    """Aggregate geo-level xr.DataArray to national level and then scale values.

    Args:
      geo_da: The geo-level DataArray to convert.
      national_da_name: The name for the returned national DataArray.
      transformer_class: The TensorTransformer class to apply after summing to
        national level. Must be None, CenteringAndScalingTransformer, or
        MediaTransformer.
      channel_dim: The name of the dimension coordinate to aggregate over (e.g.,
        constants.CONTROL_VARIABLE). If None, standard sum aggregation is used.
      da_var_agg_map: A dictionary mapping dataArray variable names to
        aggregation functions. Used only if channel_dim is not None.

    Returns:
      An xr.DataArray representing the aggregated and scaled national-level
        data.
    """
    temp_geo_dim = constants.NATIONAL_MODEL_DEFAULT_GEO_NAME

    if da_var_agg_map is None:
      da_var_agg_map = {}

    if channel_dim is not None:
      national_da = self._aggregate_variables(
          geo_da, channel_dim, da_var_agg_map
      )
    else:
      national_da = geo_da.sum(
          dim=constants.GEO, keepdims=True, skipna=False, keep_attrs=True
      )

    national_da = national_da.assign_coords({constants.GEO: [temp_geo_dim]})
    national_da.values = backend.cast(national_da.values, dtype=backend.float32)
    national_da = self._scale_xarray(national_da, transformer_class)

    national_da = national_da.sel({constants.GEO: temp_geo_dim}, drop=True)
    national_da.name = national_da_name
    return national_da

  def _get_rf_data(
      self,
      reach_raw_da: xr.DataArray,
      freq_raw_da: xr.DataArray,
      is_organic: bool,
  ) -> ReachFrequencyData:
    """Get impressions and frequencies data arrays for RF channels."""
    if is_organic:
      scaled_reach_values = (
          self._meridian.organic_rf_tensors.organic_reach_scaled
      )
      names = _ORGANIC_RF_NAMES
    else:
      scaled_reach_values = self._meridian.rf_tensors.reach_scaled
      names = _RF_NAMES

    reach_scaled_da = _data_array_like(
        da=reach_raw_da, values=scaled_reach_values
    )
    reach_scaled_da.name = names.reach_scaled
    # Truncate the media time for reach and scaled reach.
    reach_raw_da = self._truncate_media_time(reach_raw_da)
    reach_raw_da.name = names.reach
    reach_scaled_da = self._truncate_media_time(reach_scaled_da)

    # The geo level frequency
    frequency_da = self._truncate_media_time(freq_raw_da)
    frequency_da.name = names.frequency

    # The raw geo level impression
    # It's equal to reach * frequency.
    impressions_raw_da = reach_raw_da * frequency_da
    impressions_raw_da.name = names.impressions
    impressions_raw_da.values = backend.cast(
        impressions_raw_da.values, dtype=backend.float32
    )

    if self._is_national_data:
      national_reach_raw_da = reach_raw_da.squeeze(constants.GEO, drop=True)
      national_reach_raw_da.name = names.national_reach
      national_reach_scaled_da = reach_scaled_da.squeeze(
          constants.GEO, drop=True
      )
      national_reach_scaled_da.name = names.national_reach_scaled
      national_impressions_raw_da = impressions_raw_da.squeeze(
          constants.GEO, drop=True
      )
      national_impressions_raw_da.name = names.national_impressions
      national_frequency_da = frequency_da.squeeze(constants.GEO, drop=True)
      national_frequency_da.name = names.national_frequency

      # Scaled impressions
      impressions_scaled_da = self._scale_xarray(
          impressions_raw_da, transformers.MediaTransformer
      )
      impressions_scaled_da.name = names.impressions_scaled
      national_impressions_scaled_da = impressions_scaled_da.squeeze(
          constants.GEO, drop=True
      )
      national_impressions_scaled_da.name = names.national_impressions_scaled
    else:
      national_reach_raw_da = self._aggregate_and_scale_geo_da(
          reach_raw_da, names.national_reach, None
      )
      national_reach_scaled_da = self._aggregate_and_scale_geo_da(
          reach_raw_da,
          names.national_reach_scaled,
          transformers.MediaTransformer,
      )
      national_impressions_raw_da = self._aggregate_and_scale_geo_da(
          impressions_raw_da,
          names.national_impressions,
          None,
      )

      # National frequency is a weighted average of geo frequencies,
      # weighted by reach.
      national_frequency_da = xr.where(
          national_reach_raw_da == 0.0,
          0.0,
          national_impressions_raw_da / national_reach_raw_da,
      )
      national_frequency_da.name = names.national_frequency
      national_frequency_da.values = backend.cast(
          national_frequency_da.values, dtype=backend.float32
      )

      # Scale the impressions by population
      impressions_scaled_da = self._scale_xarray(
          impressions_raw_da,
          transformers.MediaTransformer,
          population=self._meridian.population,
      )
      impressions_scaled_da.name = names.impressions_scaled

      # Scale the national impressions
      national_impressions_scaled_da = self._aggregate_and_scale_geo_da(
          impressions_raw_da,
          names.national_impressions_scaled,
          transformers.MediaTransformer,
      )

    return ReachFrequencyData(
        reach_raw_da=reach_raw_da,
        reach_scaled_da=reach_scaled_da,
        national_reach_raw_da=national_reach_raw_da,
        national_reach_scaled_da=national_reach_scaled_da,
        frequency_da=frequency_da,
        national_frequency_da=national_frequency_da,
        rf_impressions_scaled_da=impressions_scaled_da,
        national_rf_impressions_scaled_da=national_impressions_scaled_da,
        rf_impressions_raw_da=impressions_raw_da,
        national_rf_impressions_raw_da=national_impressions_raw_da,
    )

  def _pairwise_corr_for_geo_data(
      self, dims: str | Sequence[str], extreme_corr_threshold: float
  ) -> tuple[xr.DataArray, pd.DataFrame]:
    """Get pairwise correlation among treatments and controls for geo data."""
    corr_mat = _compute_correlation_matrix(
        self._stacked_treatment_control_scaled_da, dims=dims
    )
    extreme_corr_var_pairs_df = _find_extreme_corr_pairs(
        corr_mat, extreme_corr_threshold
    )
    return corr_mat, extreme_corr_var_pairs_df

  def check_geo_pairwise_corr(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.PairwiseCorrArtifact]:
    """Checks pairwise correlation among treatments and controls for geo data.

    Returns:
      An EDAOutcome object with findings and result values.

    Raises:
      GeoLevelCheckOnNationalModelError: If the model is national.
    """
    # If the model is national, raise an error.
    if self._is_national_data:
      raise GeoLevelCheckOnNationalModelError(
          'check_geo_pairwise_corr is not supported for national models.'
      )

    findings = []

    overall_corr_mat, overall_extreme_corr_var_pairs_df = (
        self._pairwise_corr_for_geo_data(
            dims=[constants.GEO, constants.TIME],
            extreme_corr_threshold=_OVERALL_PAIRWISE_CORR_THRESHOLD,
        )
    )
    if not overall_extreme_corr_var_pairs_df.empty:
      var_pairs = overall_extreme_corr_var_pairs_df.index.to_list()
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ERROR,
              explanation=(
                  'Some variables have perfect pairwise correlation across all'
                  ' times and geos. For each pair of perfectly-correlated'
                  ' variables, please remove one of the variables from the'
                  f' model.\nPairs with perfect correlation: {var_pairs}'
              ),
          )
      )

    geo_corr_mat, geo_extreme_corr_var_pairs_df = (
        self._pairwise_corr_for_geo_data(
            dims=constants.TIME,
            extreme_corr_threshold=_GEO_PAIRWISE_CORR_THRESHOLD,
        )
    )
    # Overall correlation and per-geo correlation findings are mutually
    # exclusive, and overall correlation finding takes precedence.
    if (
        overall_extreme_corr_var_pairs_df.empty
        and not geo_extreme_corr_var_pairs_df.empty
    ):
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ATTENTION,
              explanation=(
                  'Some variables have perfect pairwise correlation in certain'
                  ' geo(s). Consider checking your data, and/or combining these'
                  ' variables if they also have high pairwise correlations in'
                  ' other geos.'
              ),
          )
      )

    # If there are no findings, add a INFO level finding indicating that no
    # severe correlations were found and what it means for user's data.
    if not findings:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review the computed pairwise correlations. Note that'
                  ' high pairwise correlation may cause model identifiability'
                  ' and convergence issues. Consider combining the variables if'
                  ' high correlation exists.'
              ),
          )
      )

    pairwise_corr_artifacts = [
        eda_outcome.PairwiseCorrArtifact(
            level=eda_outcome.AnalysisLevel.OVERALL,
            corr_matrix=overall_corr_mat,
            extreme_corr_var_pairs=overall_extreme_corr_var_pairs_df,
            extreme_corr_threshold=_OVERALL_PAIRWISE_CORR_THRESHOLD,
        ),
        eda_outcome.PairwiseCorrArtifact(
            level=eda_outcome.AnalysisLevel.GEO,
            corr_matrix=geo_corr_mat,
            extreme_corr_var_pairs=geo_extreme_corr_var_pairs_df,
            extreme_corr_threshold=_GEO_PAIRWISE_CORR_THRESHOLD,
        ),
    ]

    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.PAIRWISE_CORRELATION,
        findings=findings,
        analysis_artifacts=pairwise_corr_artifacts,
    )

  def check_national_pairwise_corr(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.PairwiseCorrArtifact]:
    """Checks pairwise correlation among treatments and controls for national data.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    findings = []

    corr_mat = _compute_correlation_matrix(
        self._stacked_national_treatment_control_scaled_da, dims=constants.TIME
    )
    extreme_corr_var_pairs_df = _find_extreme_corr_pairs(
        corr_mat, _NATIONAL_PAIRWISE_CORR_THRESHOLD
    )

    if not extreme_corr_var_pairs_df.empty:
      var_pairs = extreme_corr_var_pairs_df.index.to_list()
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ERROR,
              explanation=(
                  'Some variables have perfect pairwise correlation across all'
                  ' times. For each pair of perfectly-correlated'
                  ' variables, please remove one of the variables from the'
                  f' model.\nPairs with perfect correlation: {var_pairs}'
              ),
          )
      )
    else:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review the computed pairwise correlations. Note that'
                  ' high pairwise correlation may cause model identifiability'
                  ' and convergence issues. Consider combining the variables if'
                  ' high correlation exists.'
              ),
          )
      )

    pairwise_corr_artifacts = [
        eda_outcome.PairwiseCorrArtifact(
            level=eda_outcome.AnalysisLevel.NATIONAL,
            corr_matrix=corr_mat,
            extreme_corr_var_pairs=extreme_corr_var_pairs_df,
            extreme_corr_threshold=_NATIONAL_PAIRWISE_CORR_THRESHOLD,
        )
    ]
    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.PAIRWISE_CORRELATION,
        findings=findings,
        analysis_artifacts=pairwise_corr_artifacts,
    )

  def check_pairwise_corr(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.PairwiseCorrArtifact]:
    """Checks pairwise correlation among treatments and controls.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    if self._is_national_data:
      return self.check_national_pairwise_corr()

    return self.check_geo_pairwise_corr()

  def _check_std(
      self,
      data: xr.DataArray,
      level: eda_outcome.AnalysisLevel,
      zero_std_message: str,
  ) -> tuple[
      eda_outcome.EDAFinding | None, eda_outcome.StandardDeviationArtifact
  ]:
    """Helper to check standard deviation."""
    std_ds = _calculate_std(data)
    outlier_df = _calculate_outliers(data)

    finding = None
    if (std_ds[_STD_WITHOUT_OUTLIERS_VAR_NAME] < _STD_THRESHOLD).any():
      finding = eda_outcome.EDAFinding(
          severity=eda_outcome.EDASeverity.ATTENTION,
          explanation=zero_std_message,
      )

    artifact = eda_outcome.StandardDeviationArtifact(
        variable=str(data.name),
        level=level,
        std_ds=std_ds,
        outlier_df=outlier_df,
    )

    return finding, artifact

  def check_geo_std(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.StandardDeviationArtifact]:
    """Checks std for geo-level KPI, treatments, R&F, and controls."""
    if self._is_national_data:
      raise ValueError('check_geo_std is not applicable for national models.')

    findings = []
    artifacts = []

    checks = [
        (
            self.kpi_scaled_da,
            (
                'KPI has zero standard deviation after removing outliers'
                ' in certain geos, indicating weak or no signal in the response'
                ' variable for these geos.  Please review the input data,'
                ' and/or consider grouping these geos together.'
            ),
        ),
        (
            self._stacked_treatment_control_scaled_da,
            (
                'Some treatment or control variables have zero standard'
                ' deviation after removing outliers in certain geo(s). Please'
                ' review the input data. If these variables are sparse,'
                ' consider combining them to mitigate potential model'
                ' identifiability and convergence issues.'
            ),
        ),
        (
            self.all_reach_scaled_da,
            (
                'There are RF or Organic RF channels with zero variation of'
                ' reach across time at a geo after outliers are removed. If'
                ' these channels also have low variation of reach in other'
                ' geos, consider modeling them as impression-based channels'
                ' instead by taking reach * frequency.'
            ),
        ),
        (
            self.all_freq_da,
            (
                'There are RF or Organic RF channels with zero variation of'
                ' frequency across time at a geo after outliers are removed. If'
                ' these channels also have low variation of frequency in other'
                ' geos, consider modeling them as impression-based channels'
                ' instead by taking reach * frequency.'
            ),
        ),
    ]

    for data_da, message in checks:
      if data_da is None:
        continue
      finding, artifact = self._check_std(
          level=eda_outcome.AnalysisLevel.GEO,
          data=data_da,
          zero_std_message=message,
      )
      artifacts.append(artifact)
      if finding:
        findings.append(finding)

    # Add an INFO finding if no findings were added.
    if not findings:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review any identified outliers and the standard'
                  ' deviation.'
              ),
          )
      )

    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.STANDARD_DEVIATION,
        findings=findings,
        analysis_artifacts=artifacts,
    )

  def check_national_std(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.StandardDeviationArtifact]:
    """Checks std for national-level KPI, treatments, R&F, and controls."""
    findings = []
    artifacts = []

    checks = [
        (
            self.national_kpi_scaled_da,
            (
                'The standard deviation of the scaled KPI drops from positive'
                ' to zero after removing outliers, indicating sparsity of KPI'
                ' i.e. lack of signal in the response variable. Please review'
                ' the input data, and/or reconsider the feasibility of model'
                ' fitting with this dataset.'
            ),
        ),
        (
            self._stacked_national_treatment_control_scaled_da,
            (
                'The standard deviation of these scaled treatment or control'
                ' variables drops from positive to zero after removing'
                ' outliers. This indicates sparsity of these variables, which'
                ' may cause model identifiability and convergence issues.'
                ' Please review the input data, and/or consider combining these'
                ' variables to mitigate sparsity.'
            ),
        ),
        (
            self.national_all_reach_scaled_da,
            (
                'There are RF channels with totally zero variation of reach'
                ' across time at the national level after outliers are removed.'
                ' Consider modeling these RF channels as impression-based'
                ' channels instead.'
            ),
        ),
        (
            self.national_all_freq_da,
            (
                'There are RF channels with totally zero variation of frequency'
                ' across time at the national level after outliers are removed.'
                ' Consider modeling these RF channels as impression-based'
                ' channels instead.'
            ),
        ),
    ]

    for data_da, message in checks:
      if data_da is None:
        continue
      finding, artifact = self._check_std(
          data=data_da,
          level=eda_outcome.AnalysisLevel.NATIONAL,
          zero_std_message=message,
      )
      artifacts.append(artifact)
      if finding:
        findings.append(finding)

    # Add an INFO finding if no findings were added.
    if not findings:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review any identified outliers and the standard'
                  ' deviation.'
              ),
          )
      )

    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.STANDARD_DEVIATION,
        findings=findings,
        analysis_artifacts=artifacts,
    )

  def check_std(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.StandardDeviationArtifact]:
    """Checks standard deviation for treatments and controls.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    if self._is_national_data:
      return self.check_national_std()

    return self.check_geo_std()

  def check_geo_vif(self) -> eda_outcome.EDAOutcome[eda_outcome.VIFArtifact]:
    """Computes geo-level variance inflation factor among treatments and controls."""
    if self._is_national_data:
      raise ValueError(
          'Geo-level VIF checks are not applicable for national models.'
      )

    # Overall level VIF check for geo data.
    tc_da = self._stacked_treatment_control_scaled_da
    overall_threshold = self._spec.vif_spec.overall_threshold

    overall_vif_da = _calculate_vif(tc_da, _STACK_VAR_COORD_NAME)
    extreme_overall_vif_da = overall_vif_da.where(
        overall_vif_da > overall_threshold
    )
    extreme_overall_vif_df = extreme_overall_vif_da.to_dataframe(
        name=_VIF_COL_NAME
    ).dropna()

    overall_vif_artifact = eda_outcome.VIFArtifact(
        level=eda_outcome.AnalysisLevel.OVERALL,
        vif_da=overall_vif_da,
        outlier_df=extreme_overall_vif_df,
    )

    # Geo level VIF check.
    geo_threshold = self._spec.vif_spec.geo_threshold
    geo_vif_da = tc_da.groupby(constants.GEO).map(
        lambda x: _calculate_vif(x, _STACK_VAR_COORD_NAME)
    )
    extreme_geo_vif_da = geo_vif_da.where(geo_vif_da > geo_threshold)
    extreme_geo_vif_df = extreme_geo_vif_da.to_dataframe(
        name=_VIF_COL_NAME
    ).dropna()

    geo_vif_artifact = eda_outcome.VIFArtifact(
        level=eda_outcome.AnalysisLevel.GEO,
        vif_da=geo_vif_da,
        outlier_df=extreme_geo_vif_df,
    )

    findings = []
    if not extreme_overall_vif_df.empty:
      high_vif_vars = extreme_overall_vif_df.index.to_list()
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ERROR,
              explanation=(
                  'Some variables have extreme multicollinearity (VIF'
                  f' >{overall_threshold}) across all times and geos. To'
                  ' address multicollinearity, please drop any variable that'
                  ' is a linear combination of other variables. Otherwise,'
                  ' consider combining variables.\n'
                  f'Variables with extreme VIF: {high_vif_vars}'
              ),
          )
      )
    elif not extreme_geo_vif_df.empty:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ATTENTION,
              explanation=(
                  'Some variables have extreme multicollinearity (with VIF >'
                  f' {geo_threshold}) in certain geo(s). Consider checking your'
                  ' data, and/or combining these variables if they also have'
                  ' high VIF in other geos.'
              ),
          )
      )
    else:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review the computed VIFs. Note that high VIF suggests'
                  ' multicollinearity issues in the dataset, which may'
                  ' jeopardize model identifiability and model convergence.'
                  ' Consider combining the variables if high VIF occurs.'
              ),
          )
      )

    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.MULTICOLLINEARITY,
        findings=findings,
        analysis_artifacts=[overall_vif_artifact, geo_vif_artifact],
    )

  def check_national_vif(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.VIFArtifact]:
    """Computes national-level variance inflation factor among treatments and controls."""
    national_tc_da = self._stacked_national_treatment_control_scaled_da
    national_threshold = self._spec.vif_spec.national_threshold
    national_vif_da = _calculate_vif(national_tc_da, _STACK_VAR_COORD_NAME)

    extreme_national_vif_df = (
        national_vif_da.where(national_vif_da > national_threshold)
        .to_dataframe(name=_VIF_COL_NAME)
        .dropna()
    )
    national_vif_artifact = eda_outcome.VIFArtifact(
        level=eda_outcome.AnalysisLevel.NATIONAL,
        vif_da=national_vif_da,
        outlier_df=extreme_national_vif_df,
    )

    findings = []
    if not extreme_national_vif_df.empty:
      high_vif_vars = extreme_national_vif_df.index.to_list()
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.ERROR,
              explanation=(
                  'Some variables have extreme multicollinearity (with VIF >'
                  f' {national_threshold}) across all times. To address'
                  ' multicollinearity, please drop any variable that is a'
                  ' linear combination of other variables. Otherwise, consider'
                  ' combining variables.\n'
                  f'Variables with extreme VIF: {high_vif_vars}'
              ),
          )
      )
    else:
      findings.append(
          eda_outcome.EDAFinding(
              severity=eda_outcome.EDASeverity.INFO,
              explanation=(
                  'Please review the computed VIFs. Note that high VIF suggests'
                  ' multicollinearity issues in the dataset, which may'
                  ' jeopardize model identifiability and model convergence.'
                  ' Consider combining the variables if high VIF occurs.'
              ),
          )
      )
    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.MULTICOLLINEARITY,
        findings=findings,
        analysis_artifacts=[national_vif_artifact],
    )

  def check_vif(self) -> eda_outcome.EDAOutcome[eda_outcome.VIFArtifact]:
    """Computes variance inflation factor among treatments and controls.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    if self._is_national_data:
      return self.check_national_vif()

    return self.check_geo_vif()

  @property
  def kpi_has_variability(self) -> bool:
    """Returns True if the KPI has variability across geos and times."""
    return (
        self._overall_scaled_kpi_invariability_artifact.kpi_stdev.item()
        >= _STD_THRESHOLD
    )

  def check_overall_kpi_invariability(self) -> eda_outcome.EDAOutcome:
    """Checks if the KPI is constant across all geos and times."""
    kpi = self._overall_scaled_kpi_invariability_artifact.kpi_da.name
    geo_text = '' if self._is_national_data else 'geos and '

    if not self.kpi_has_variability:
      eda_finding = eda_outcome.EDAFinding(
          severity=eda_outcome.EDASeverity.ERROR,
          explanation=(
              f'`{kpi}` is constant across all {geo_text}times, indicating no'
              ' signal in the data. Please fix this data error.'
          ),
      )
    else:
      eda_finding = eda_outcome.EDAFinding(
          severity=eda_outcome.EDASeverity.INFO,
          explanation=(
              f'The {kpi} has variability across {geo_text}times in the data.'
          ),
      )

    return eda_outcome.EDAOutcome(
        check_type=eda_outcome.EDACheckType.KPI_INVARIABILITY,
        findings=[eda_finding],
        analysis_artifacts=[self._overall_scaled_kpi_invariability_artifact],
    )

  def check_geo_cost_per_media_unit(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.CostPerMediaUnitArtifact]:
    """Checks if the cost per media unit is valid for geo data.

    Returns:
      An EDAOutcome object with findings and result values.

    Raises:
      GeoLevelCheckOnNationalModelError: If the check is called for a national
        model.
    """
    if self._is_national_data:
      raise GeoLevelCheckOnNationalModelError(
          'check_geo_cost_per_media_unit is not supported for national models.'
      )
    return _check_cost_per_media_unit(
        self.all_spend_ds,
        self.paid_raw_media_units_ds,
        eda_outcome.AnalysisLevel.GEO,
    )

  def check_national_cost_per_media_unit(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.CostPerMediaUnitArtifact]:
    """Checks if the cost per media unit is valid for national data.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    return _check_cost_per_media_unit(
        self.national_all_spend_ds,
        self.national_paid_raw_media_units_ds,
        eda_outcome.AnalysisLevel.NATIONAL,
    )

  def check_cost_per_media_unit(
      self,
  ) -> eda_outcome.EDAOutcome[eda_outcome.CostPerMediaUnitArtifact]:
    """Checks if the cost per media unit is valid.

    This function checks the following conditions:
    1. cost == 0 and media unit > 0.
    2. cost > 0 and media unit == 0.
    3. cost_per_media_unit has outliers.

    Returns:
      An EDAOutcome object with findings and result values.
    """
    if self._is_national_data:
      return self.check_national_cost_per_media_unit()

    return self.check_geo_cost_per_media_unit()

  def run_all_critical_checks(self) -> list[eda_outcome.EDAOutcome]:
    """Runs all critical EDA checks.

    Critical checks are those that can result in EDASeverity.ERROR findings.

    Returns:
      A list of EDA outcomes, one for each check.
    """
    outcomes = []
    for check, check_type in self._critical_checks:
      try:
        outcomes.append(check())
      except Exception as e:  # pylint: disable=broad-except
        error_finding = eda_outcome.EDAFinding(
            severity=eda_outcome.EDASeverity.ERROR,
            explanation=(
                f'An error occurred during running {check.__name__}: {e!r}'
            ),
        )
        outcomes.append(
            eda_outcome.EDAOutcome(
                check_type=check_type,
                findings=[error_finding],
                analysis_artifacts=[],
            )
        )
    return outcomes



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines ModelContext class for Meridian."""

from collections.abc import Mapping, Sequence
import functools
import warnings

from meridian import backend
from meridian import constants
from meridian.data import input_data as data
from meridian.model import adstock_hill
from meridian.model import knots
from meridian.model import media
from meridian.model import prior_distribution
from meridian.model import spec
from meridian.model import transformers
import numpy as np

__all__ = [
    "ModelContext",
]


class ModelContext:
  """Model context for Meridian.

  This class contains all model parameters that do not change between the runs
  of Meridian.
  """

  def __init__(
      self,
      input_data: data.InputData,
      model_spec: spec.ModelSpec,
  ):
    self._input_data = input_data
    self._model_spec = model_spec

    self._validate_data_dependent_model_spec()
    self._validate_model_spec_shapes()

    self._set_total_media_contribution_prior = False
    self._warn_setting_ignored_priors()
    self._validate_mroi_priors_non_revenue()
    self._validate_roi_priors_non_revenue()
    self._check_media_prior_support()
    self._validate_geo_invariants()
    self._validate_time_invariants()

  def _validate_data_dependent_model_spec(self):
    """Validates that the data dependent model specs have correct shapes."""

    if self._model_spec.roi_calibration_period is not None and (
        self._model_spec.roi_calibration_period.shape
        != (
            self.n_media_times,
            self.n_media_channels,
        )
    ):
      raise ValueError(
          "The shape of `roi_calibration_period`"
          f" {self._model_spec.roi_calibration_period.shape} is different from"
          f" `(n_media_times, n_media_channels) = ({self.n_media_times},"
          f" {self.n_media_channels})`."
      )

    if self._model_spec.rf_roi_calibration_period is not None and (
        self._model_spec.rf_roi_calibration_period.shape
        != (
            self.n_media_times,
            self.n_rf_channels,
        )
    ):
      raise ValueError(
          "The shape of `rf_roi_calibration_period`"
          f" {self._model_spec.rf_roi_calibration_period.shape} is different"
          f" from `(n_media_times, n_rf_channels) = ({self.n_media_times},"
          f" {self.n_rf_channels})`."
      )

    if self._model_spec.holdout_id is not None:
      if self.is_national and (
          self._model_spec.holdout_id.shape != (self.n_times,)
      ):
        raise ValueError(
            f"The shape of `holdout_id` {self._model_spec.holdout_id.shape} is"
            f" different from `(n_times,) = ({self.n_times},)`."
        )
      elif not self.is_national and (
          self._model_spec.holdout_id.shape
          != (
              self.n_geos,
              self.n_times,
          )
      ):
        raise ValueError(
            f"The shape of `holdout_id` {self._model_spec.holdout_id.shape} is"
            f" different from `(n_geos, n_times) = ({self.n_geos},"
            f" {self.n_times})`."
        )

    if self._model_spec.control_population_scaling_id is not None and (
        self._model_spec.control_population_scaling_id.shape
        != (self.n_controls,)
    ):
      raise ValueError(
          "The shape of `control_population_scaling_id`"
          f" {self._model_spec.control_population_scaling_id.shape} is"
          f" different from `(n_controls,) = ({self.n_controls},)`."
      )

    if self._model_spec.non_media_population_scaling_id is not None and (
        self._model_spec.non_media_population_scaling_id.shape
        != (self.n_non_media_channels,)
    ):
      raise ValueError(
          "The shape of `non_media_population_scaling_id`"
          f" {self._model_spec.non_media_population_scaling_id.shape} is"
          " different from `(n_non_media_channels,) ="
          f" ({self.n_non_media_channels},)`."
      )

  def _validate_model_spec_shapes(self):
    """Validate shapes of model_spec attributes."""
    if self._model_spec.roi_calibration_period is not None:
      if self._model_spec.roi_calibration_period.shape != (
          self.n_media_times,
          self.n_media_channels,
      ):
        raise ValueError(
            "The shape of `roi_calibration_period`"
            f" {self._model_spec.roi_calibration_period.shape} is different"
            f" from `(n_media_times, n_media_channels) = ({self.n_media_times},"
            f" {self.n_media_channels})`."
        )

    if self._model_spec.rf_roi_calibration_period is not None:
      if self._model_spec.rf_roi_calibration_period.shape != (
          self.n_media_times,
          self.n_rf_channels,
      ):
        raise ValueError(
            "The shape of `rf_roi_calibration_period`"
            f" {self._model_spec.rf_roi_calibration_period.shape} is different"
            f" from `(n_media_times, n_rf_channels) = ({self.n_media_times},"
            f" {self.n_rf_channels})`."
        )

    if self._model_spec.holdout_id is not None:
      expected_shape = (
          (self.n_times,) if self.is_national else (self.n_geos, self.n_times)
      )
      if self._model_spec.holdout_id.shape != expected_shape:
        raise ValueError(
            f"The shape of `holdout_id` {self._model_spec.holdout_id.shape} is"
            " different from"
            f" {'`(n_times,)`' if self.is_national else '`(n_geos, n_times)`'} ="
            f" {expected_shape}."
        )

    if self._model_spec.control_population_scaling_id is not None:
      if self._model_spec.control_population_scaling_id.shape != (
          self.n_controls,
      ):
        raise ValueError(
            "The shape of `control_population_scaling_id`"
            f" {self._model_spec.control_population_scaling_id.shape} is"
            f" different from `(n_controls,) = ({self.n_controls},)`."
        )

  def _validate_geo_invariants(self):
    """Validates non-national model invariants."""
    if self.is_national:
      return

    if self._input_data.controls is not None:
      self._check_if_no_geo_variation(
          self.controls_scaled,
          constants.CONTROLS,
          self._input_data.controls.coords[constants.CONTROL_VARIABLE].values,
      )
    if self._input_data.non_media_treatments is not None:
      self._check_if_no_geo_variation(
          self.non_media_treatments_normalized,
          constants.NON_MEDIA_TREATMENTS,
          self._input_data.non_media_treatments.coords[
              constants.NON_MEDIA_CHANNEL
          ].values,
      )
    if self._input_data.media is not None:
      self._check_if_no_geo_variation(
          self.media_tensors.media_scaled,
          constants.MEDIA,
          self._input_data.media.coords[constants.MEDIA_CHANNEL].values,
      )
    if self._input_data.reach is not None:
      self._check_if_no_geo_variation(
          self.rf_tensors.reach_scaled,
          constants.REACH,
          self._input_data.reach.coords[constants.RF_CHANNEL].values,
      )
    if self._input_data.organic_media is not None:
      self._check_if_no_geo_variation(
          self.organic_media_tensors.organic_media_scaled,
          "organic_media",
          self._input_data.organic_media.coords[
              constants.ORGANIC_MEDIA_CHANNEL
          ].values,
      )
    if self._input_data.organic_reach is not None:
      self._check_if_no_geo_variation(
          self.organic_rf_tensors.organic_reach_scaled,
          "organic_reach",
          self._input_data.organic_reach.coords[
              constants.ORGANIC_RF_CHANNEL
          ].values,
      )

  def _check_if_no_geo_variation(
      self,
      scaled_data: backend.Tensor,
      data_name: str,
      data_dims: Sequence[str],
      epsilon=1e-4,
  ):
    """Raise an error if `n_knots == n_time` and data lacks geo variation."""

    # Result shape: [n, d], where d is the number of axes of condition.
    col_idx_full = backend.get_indices_where(
        backend.reduce_std(scaled_data, axis=0) < epsilon
    )[:, 1]
    col_idx_unique, _, counts = backend.unique_with_counts(col_idx_full)
    # We use the shape of scaled_data (instead of `n_time`) because the data may
    # be padded to account for lagged effects.
    data_n_time = scaled_data.shape[1]
    mask = backend.equal(counts, data_n_time)
    col_idx_bad = backend.boolean_mask(col_idx_unique, mask)
    dims_bad = backend.gather(data_dims, col_idx_bad)

    if col_idx_bad.shape[0] and self.knot_info.n_knots == self.n_times:
      raise ValueError(
          f"The following {data_name} variables do not vary across geos, making"
          f" a model with n_knots=n_time unidentifiable: {dims_bad}. This can"
          " lead to poor model convergence. Since these variables only vary"
          " across time and not across geo, they are collinear with time and"
          " redundant in a model with a parameter for each time period.  To"
          " address this, you can either: (1) decrease the number of knots"
          " (n_knots < n_time), or (2) drop the listed variables that do not"
          " vary across geos."
      )

  def _validate_time_invariants(self):
    """Validates model time invariants."""
    if self._input_data.controls is not None:
      self._check_if_no_time_variation(
          self.controls_scaled,
          constants.CONTROLS,
          self._input_data.controls.coords[constants.CONTROL_VARIABLE].values,
      )
    if self._input_data.non_media_treatments is not None:
      self._check_if_no_time_variation(
          self.non_media_treatments_normalized,
          constants.NON_MEDIA_TREATMENTS,
          self._input_data.non_media_treatments.coords[
              constants.NON_MEDIA_CHANNEL
          ].values,
      )
    if self._input_data.media is not None:
      self._check_if_no_time_variation(
          self.media_tensors.media_scaled,
          constants.MEDIA,
          self._input_data.media.coords[constants.MEDIA_CHANNEL].values,
      )
    if self._input_data.reach is not None:
      self._check_if_no_time_variation(
          self.rf_tensors.reach_scaled,
          constants.REACH,
          self._input_data.reach.coords[constants.RF_CHANNEL].values,
      )
    if self._input_data.organic_media is not None:
      self._check_if_no_time_variation(
          self.organic_media_tensors.organic_media_scaled,
          constants.ORGANIC_MEDIA,
          self._input_data.organic_media.coords[
              constants.ORGANIC_MEDIA_CHANNEL
          ].values,
      )
    if self._input_data.organic_reach is not None:
      self._check_if_no_time_variation(
          self.organic_rf_tensors.organic_reach_scaled,
          constants.ORGANIC_REACH,
          self._input_data.organic_reach.coords[
              constants.ORGANIC_RF_CHANNEL
          ].values,
      )

  def _check_if_no_time_variation(
      self,
      scaled_data: backend.Tensor,
      data_name: str,
      data_dims: Sequence[str],
      epsilon=1e-4,
  ):
    """Raise an error if data lacks time variation."""

    # Result shape: [n, d], where d is the number of axes of condition.
    col_idx_full = backend.get_indices_where(
        backend.reduce_std(scaled_data, axis=1) < epsilon
    )[:, 1]
    col_idx_unique, _, counts = backend.unique_with_counts(col_idx_full)
    mask = backend.equal(counts, self.n_geos)
    col_idx_bad = backend.boolean_mask(col_idx_unique, mask)
    dims_bad = backend.gather(data_dims, col_idx_bad)
    if col_idx_bad.shape[0]:
      if self.is_national:
        raise ValueError(
            f"The following {data_name} variables do not vary across time,"
            " which is equivalent to no signal at all in a national model:"
            f" {dims_bad}.  This can lead to poor model convergence. To address"
            " this, drop the listed variables that do not vary across time."
        )
      else:
        raise ValueError(
            f"The following {data_name} variables do not vary across time,"
            f" making a model with geo main effects unidentifiable: {dims_bad}."
            " This can lead to poor model convergence. Since these variables"
            " only vary across geo and not across time, they are collinear"
            " with geo and redundant in a model with geo main effects. To"
            " address this, drop the listed variables that do not vary across"
            " time."
        )

  @property
  def input_data(self) -> data.InputData:
    return self._input_data

  @property
  def model_spec(self) -> spec.ModelSpec:
    return self._model_spec

  @functools.cached_property
  def media_tensors(self) -> media.MediaTensors:
    return media.build_media_tensors(self._input_data, self._model_spec)

  @functools.cached_property
  def rf_tensors(self) -> media.RfTensors:
    return media.build_rf_tensors(self._input_data, self._model_spec)

  @functools.cached_property
  def organic_media_tensors(self) -> media.OrganicMediaTensors:
    return media.build_organic_media_tensors(self._input_data)

  @functools.cached_property
  def organic_rf_tensors(self) -> media.OrganicRfTensors:
    return media.build_organic_rf_tensors(self._input_data)

  @functools.cached_property
  def kpi(self) -> backend.Tensor:
    return backend.to_tensor(self._input_data.kpi, dtype=backend.float32)

  @functools.cached_property
  def revenue_per_kpi(self) -> backend.Tensor | None:
    if self._input_data.revenue_per_kpi is None:
      return None
    return backend.to_tensor(
        self._input_data.revenue_per_kpi, dtype=backend.float32
    )

  @functools.cached_property
  def controls(self) -> backend.Tensor | None:
    if self._input_data.controls is None:
      return None
    return backend.to_tensor(self._input_data.controls, dtype=backend.float32)

  @functools.cached_property
  def non_media_treatments(self) -> backend.Tensor | None:
    if self._input_data.non_media_treatments is None:
      return None
    return backend.to_tensor(
        self._input_data.non_media_treatments, dtype=backend.float32
    )

  @functools.cached_property
  def population(self) -> backend.Tensor:
    return backend.to_tensor(self._input_data.population, dtype=backend.float32)

  @functools.cached_property
  def total_spend(self) -> backend.Tensor:
    return backend.to_tensor(
        self._input_data.get_total_spend(), dtype=backend.float32
    )

  @functools.cached_property
  def total_outcome(self) -> backend.Tensor:
    return backend.to_tensor(
        self._input_data.get_total_outcome(), dtype=backend.float32
    )

  @property
  def n_geos(self) -> int:
    return len(self._input_data.geo)

  @property
  def n_media_channels(self) -> int:
    if self._input_data.media_channel is None:
      return 0
    return len(self._input_data.media_channel)

  @property
  def n_rf_channels(self) -> int:
    if self._input_data.rf_channel is None:
      return 0
    return len(self._input_data.rf_channel)

  @property
  def n_organic_media_channels(self) -> int:
    if self._input_data.organic_media_channel is None:
      return 0
    return len(self._input_data.organic_media_channel)

  @property
  def n_organic_rf_channels(self) -> int:
    if self._input_data.organic_rf_channel is None:
      return 0
    return len(self._input_data.organic_rf_channel)

  @property
  def n_controls(self) -> int:
    if self._input_data.control_variable is None:
      return 0
    return len(self._input_data.control_variable)

  @property
  def n_non_media_channels(self) -> int:
    if self._input_data.non_media_channel is None:
      return 0
    return len(self._input_data.non_media_channel)

  @property
  def n_times(self) -> int:
    return len(self._input_data.time)

  @property
  def n_media_times(self) -> int:
    return len(self._input_data.media_time)

  @property
  def is_national(self) -> bool:
    return self.n_geos == 1

  @functools.cached_property
  def knot_info(self) -> knots.KnotInfo:
    return knots.get_knot_info(
        n_times=self.n_times,
        knots=self._model_spec.knots,
        enable_aks=self._model_spec.enable_aks,
        data=self._input_data,
        is_national=self.is_national,
    )

  @functools.cached_property
  def controls_transformer(
      self,
  ) -> transformers.CenteringAndScalingTransformer | None:
    """Returns a `CenteringAndScalingTransformer` for controls, if it exists."""
    if self.controls is None:
      return None

    if self._model_spec.control_population_scaling_id is not None:
      controls_population_scaling_id = backend.to_tensor(
          self._model_spec.control_population_scaling_id, dtype=backend.bool_
      )
    else:
      controls_population_scaling_id = None

    return transformers.CenteringAndScalingTransformer(
        tensor=self.controls,
        population=self.population,
        population_scaling_id=controls_population_scaling_id,
    )

  @functools.cached_property
  def non_media_transformer(
      self,
  ) -> transformers.CenteringAndScalingTransformer | None:
    """Returns a `CenteringAndScalingTransformer` for non-media treatments."""
    if self.non_media_treatments is None:
      return None
    if self._model_spec.non_media_population_scaling_id is not None:
      non_media_population_scaling_id = backend.to_tensor(
          self._model_spec.non_media_population_scaling_id, dtype=backend.bool_
      )
    else:
      non_media_population_scaling_id = None

    return transformers.CenteringAndScalingTransformer(
        tensor=self.non_media_treatments,
        population=self.population,
        population_scaling_id=non_media_population_scaling_id,
    )

  @functools.cached_property
  def kpi_transformer(self) -> transformers.KpiTransformer:
    return transformers.KpiTransformer(self.kpi, self.population)

  @functools.cached_property
  def controls_scaled(self) -> backend.Tensor | None:
    if self.controls is not None:
      # If `controls` is defined, then `controls_transformer` is also defined.
      return self.controls_transformer.forward(self.controls)  # pytype: disable=attribute-error
    else:
      return None

  @functools.cached_property
  def non_media_treatments_normalized(self) -> backend.Tensor | None:
    """Normalized non-media treatments.

    The non-media treatments values are scaled by population (for channels where
    `non_media_population_scaling_id` is `True`) and normalized by centering and
    scaling with means and standard deviations.
    """
    if self.non_media_transformer is not None:
      return self.non_media_transformer.forward(
          self.non_media_treatments
      )  # pytype: disable=attribute-error
    else:
      return None

  @functools.cached_property
  def kpi_scaled(self) -> backend.Tensor:
    return self.kpi_transformer.forward(self.kpi)

  @functools.cached_property
  def media_effects_dist(self) -> str:
    if self.is_national:
      return constants.NATIONAL_MODEL_SPEC_ARGS[constants.MEDIA_EFFECTS_DIST]
    else:
      return self._model_spec.media_effects_dist

  @functools.cached_property
  def unique_sigma_for_each_geo(self) -> bool:
    if self.is_national:
      # Should evaluate to False.
      return constants.NATIONAL_MODEL_SPEC_ARGS[
          constants.UNIQUE_SIGMA_FOR_EACH_GEO
      ]
    else:
      return self._model_spec.unique_sigma_for_each_geo

  @functools.cached_property
  def baseline_geo_idx(self) -> int:
    """Returns the index of the baseline geo."""
    if isinstance(self._model_spec.baseline_geo, int):
      if (
          self._model_spec.baseline_geo < 0
          or self._model_spec.baseline_geo >= self.n_geos
      ):
        raise ValueError(
            f"Baseline geo index {self._model_spec.baseline_geo} out of range"
            f" [0, {self.n_geos - 1}]."
        )
      return self._model_spec.baseline_geo
    elif isinstance(self._model_spec.baseline_geo, str):
      # np.where returns a 1-D tuple, its first element is an array of found
      # elements.
      index = np.where(self._input_data.geo == self._model_spec.baseline_geo)[0]
      if index.size == 0:
        raise ValueError(
            f"Baseline geo '{self._model_spec.baseline_geo}' not found."
        )
      # Geos are unique, so index is a 1-element array.
      return index[0]
    else:
      return backend.argmax(self.population)

  @functools.cached_property
  def holdout_id(self) -> backend.Tensor | None:
    if self._model_spec.holdout_id is None:
      return None
    tensor = backend.to_tensor(self._model_spec.holdout_id, dtype=backend.bool_)
    return tensor[backend.newaxis, ...] if self.is_national else tensor

  def _warn_setting_ignored_priors(self):
    """Raises a warning if ignored priors are set."""
    default_distribution = prior_distribution.PriorDistribution()
    for ignored_priors_dict, prior_type, prior_type_name in (
        (
            constants.IGNORED_PRIORS_MEDIA,
            self._model_spec.effective_media_prior_type,
            "media_prior_type",
        ),
        (
            constants.IGNORED_PRIORS_RF,
            self._model_spec.effective_rf_prior_type,
            "rf_prior_type",
        ),
    ):
      ignored_custom_priors = []
      for prior in ignored_priors_dict.get(prior_type, []):
        self_prior = getattr(self._model_spec.prior, prior)
        default_prior = getattr(default_distribution, prior)
        if not prior_distribution.distributions_are_equal(
            self_prior, default_prior
        ):
          ignored_custom_priors.append(prior)
      if ignored_custom_priors:
        ignored_priors_str = ", ".join(ignored_custom_priors)
        warnings.warn(
            f"Custom prior(s) `{ignored_priors_str}` are ignored when"
            f' `{prior_type_name}` is set to "{prior_type}".'
        )

  def _validate_mroi_priors_non_revenue(self):
    """Validates mroi priors in the non-revenue outcome case."""
    if (
        self._input_data.kpi_type == constants.NON_REVENUE
        and self._input_data.revenue_per_kpi is None
    ):
      default_distribution = prior_distribution.PriorDistribution()
      if (
          self.n_media_channels > 0
          and (
              self._model_spec.effective_media_prior_type
              == constants.TREATMENT_PRIOR_TYPE_MROI
          )
          and prior_distribution.distributions_are_equal(
              self._model_spec.prior.mroi_m, default_distribution.mroi_m
          )
      ):
        raise ValueError(
            f"Custom priors should be set on `{constants.MROI_M}` when"
            ' `media_prior_type` is "mroi", KPI is non-revenue and revenue per'
            " kpi data is missing."
        )
      if (
          self.n_rf_channels > 0
          and (
              self._model_spec.effective_rf_prior_type
              == constants.TREATMENT_PRIOR_TYPE_MROI
          )
          and prior_distribution.distributions_are_equal(
              self._model_spec.prior.mroi_rf, default_distribution.mroi_rf
          )
      ):
        raise ValueError(
            f"Custom priors should be set on `{constants.MROI_RF}` when"
            ' `rf_prior_type` is "mroi", KPI is non-revenue and revenue per kpi'
            " data is missing."
        )

  def _validate_roi_priors_non_revenue(self):
    """Validates roi priors in the non-revenue outcome case."""
    if (
        self._input_data.kpi_type == constants.NON_REVENUE
        and self._input_data.revenue_per_kpi is None
    ):
      default_distribution = prior_distribution.PriorDistribution()
      default_roi_m_used = (
          self._model_spec.effective_media_prior_type
          == constants.TREATMENT_PRIOR_TYPE_ROI
          and prior_distribution.distributions_are_equal(
              self._model_spec.prior.roi_m, default_distribution.roi_m
          )
      )
      default_roi_rf_used = (
          self._model_spec.effective_rf_prior_type
          == constants.TREATMENT_PRIOR_TYPE_ROI
          and prior_distribution.distributions_are_equal(
              self._model_spec.prior.roi_rf, default_distribution.roi_rf
          )
      )
      # If ROI priors are used with the default prior distribution for all paid
      # channels (media and RF), then use the "total paid media contribution
      # prior" procedure.
      if (
          (default_roi_m_used and default_roi_rf_used)
          or (self.n_media_channels == 0 and default_roi_rf_used)
          or (self.n_rf_channels == 0 and default_roi_m_used)
      ):
        self._set_total_media_contribution_prior = True
        warnings.warn(
            "Consider setting custom ROI priors, as kpi_type was specified as"
            " `non_revenue` with no `revenue_per_kpi` being set. Otherwise, the"
            " total media contribution prior will be used with"
            f" `p_mean={constants.P_MEAN}` and `p_sd={constants.P_SD}`. Further"
            " documentation available at "
            " https://developers.google.com/meridian/docs/advanced-modeling/unknown-revenue-kpi-custom#set-total-paid-media-contribution-prior",
        )
      elif self.n_media_channels > 0 and default_roi_m_used:
        raise ValueError(
            f"Custom priors should be set on `{constants.ROI_M}` when"
            ' `media_prior_type` is "roi", custom priors are assigned on'
            ' `{constants.ROI_RF}` or `rf_prior_type` is not "roi", KPI is'
            " non-revenue and revenue per kpi data is missing."
        )
      elif self.n_rf_channels > 0 and default_roi_rf_used:
        raise ValueError(
            f"Custom priors should be set on `{constants.ROI_RF}` when"
            ' `rf_prior_type` is "roi", custom priors are assigned on'
            ' `{constants.ROI_M}` or `media_prior_type` is not "roi", KPI is'
            " non-revenue and revenue per kpi data is missing."
        )

  def _check_media_prior_support(self) -> None:
    """Checks ROI, mROI, and Contribution prior support when random effects are log-normal.

    Priors for ROI, mROI, and Contribution can only have negative support if the
    random effects follow a normal distribution. This check enforces that priors
    have non-negative support when random effects follow a log-normal
    distribution. This check only applies to geo-level models with log-normal
    random effects since national models do not have random effects.
    """
    prior = self._model_spec.prior
    if self.n_media_channels > 0:
      self._check_for_negative_support(
          prior.roi_m,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_ROI,
      )
      self._check_for_negative_support(
          prior.mroi_m,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_MROI,
      )
      self._check_for_negative_support(
          prior.contribution_m,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      )
    if self.n_rf_channels > 0:
      self._check_for_negative_support(
          prior.roi_rf,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_ROI,
      )
      self._check_for_negative_support(
          prior.mroi_rf,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_MROI,
      )
      self._check_for_negative_support(
          prior.contribution_rf,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      )
    if self.n_organic_media_channels > 0:
      self._check_for_negative_support(
          prior.contribution_om,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      )
    if self.n_organic_rf_channels > 0:
      self._check_for_negative_support(
          prior.contribution_orf,
          self.media_effects_dist,
          constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION,
      )

  def _check_for_negative_support(
      self,
      dist: backend.tfd.Distribution,
      media_effects_dist: str,
      prior_type: str,
  ) -> None:
    """Checks for negative support in prior distributions.

    When `media_effects_dist` is `MEDIA_EFFECTS_LOG_NORMAL`, prior distributions
    for media effects must be non-negative. This function raises a ValueError if
    any part of the distribution's CDF is greater than 0 at 0, indicating some
    probability mass below zero.

    Args:
      dist: The distribution to check.
      media_effects_dist: The type of media effects distribution.
      prior_type: The prior type that corresponds with current prior under test.

    Raises:
      ValueError: If the prior distribution has negative support when
      `media_effects_dist` is `MEDIA_EFFECTS_LOG_NORMAL`.
    """
    if (
        prior_type == self._model_spec.media_prior_type
        and media_effects_dist == constants.MEDIA_EFFECTS_LOG_NORMAL
        and np.any(dist.cdf(0) > 0)
    ):
      raise ValueError(
          "Media priors must have non-negative support when"
          f' `media_effects_dist`="{media_effects_dist}". Found negative prior'
          f" distribution support for {dist.name}."
      )

  @functools.cached_property
  def prior_broadcast(self) -> prior_distribution.PriorDistribution:
    """Returns broadcasted `PriorDistribution` object."""
    total_spend = self._input_data.get_total_spend()
    # Total spend can have 1, 2 or 3 dimensions. Aggregate by channel.
    if len(total_spend.shape) == 1:
      # Already aggregated by channel.
      agg_total_spend = total_spend
    elif len(total_spend.shape) == 2:
      agg_total_spend = np.sum(total_spend, axis=(0,))
    else:
      agg_total_spend = np.sum(total_spend, axis=(0, 1))

    return self._model_spec.prior.broadcast(
        n_geos=self.n_geos,
        n_media_channels=self.n_media_channels,
        n_rf_channels=self.n_rf_channels,
        n_organic_media_channels=self.n_organic_media_channels,
        n_organic_rf_channels=self.n_organic_rf_channels,
        n_controls=self.n_controls,
        n_non_media_channels=self.n_non_media_channels,
        unique_sigma_for_each_geo=self.unique_sigma_for_each_geo,
        n_knots=self.knot_info.n_knots,
        is_national=self.is_national,
        set_total_media_contribution_prior=self._set_total_media_contribution_prior,
        kpi=np.sum(self._input_data.kpi.values),
        total_spend=agg_total_spend,
    )

  @functools.cached_property
  def adstock_decay_spec(self) -> adstock_hill.AdstockDecaySpec:
    """Returns `AdstockDecaySpec` object with correctly mapped channels."""
    if isinstance(self._model_spec.adstock_decay_spec, str):
      return adstock_hill.AdstockDecaySpec.from_consistent_type(
          self._model_spec.adstock_decay_spec
      )

    try:
      return self._create_adstock_decay_functions_from_channel_map(
          self._model_spec.adstock_decay_spec
      )
    except KeyError as e:
      raise ValueError(
          "Unrecognized channel names found in `adstock_decay_spec` keys"
          f" {tuple(self._model_spec.adstock_decay_spec.keys())}. Keys should"
          " either contain only channel_names"
          f" {tuple(self._input_data.get_all_adstock_hill_channels().tolist())} or"
          " be one or more of {'media', 'rf', 'organic_media',"
          " 'organic_rf'}."
      ) from e

  def _create_adstock_decay_functions_from_channel_map(
      self, channel_function_map: Mapping[str, str]
  ) -> adstock_hill.AdstockDecaySpec:
    """Create `AdstockDecaySpec` from mapping from channels to decay functions."""

    for channel in channel_function_map:
      if channel not in self._input_data.get_all_adstock_hill_channels():
        raise KeyError(f"Channel {channel} not found in data.")

    if self._input_data.media_channel is not None:
      media_channel_builder = self._input_data.get_paid_media_channels_argument_builder().with_default_value(
          constants.GEOMETRIC_DECAY
      )
      media_adstock_function = media_channel_builder(**channel_function_map)
    else:
      media_adstock_function = constants.GEOMETRIC_DECAY

    if self._input_data.rf_channel is not None:
      rf_channel_builder = self._input_data.get_paid_rf_channels_argument_builder().with_default_value(
          constants.GEOMETRIC_DECAY
      )
      rf_adstock_function = rf_channel_builder(**channel_function_map)
    else:
      rf_adstock_function = constants.GEOMETRIC_DECAY

    if self._input_data.organic_media_channel is not None:
      organic_media_channel_builder = self._input_data.get_organic_media_channels_argument_builder().with_default_value(
          constants.GEOMETRIC_DECAY
      )
      organic_media_adstock_function = organic_media_channel_builder(
          **channel_function_map
      )
    else:
      organic_media_adstock_function = constants.GEOMETRIC_DECAY

    if self._input_data.organic_rf_channel is not None:
      organic_rf_channel_builder = self._input_data.get_organic_rf_channels_argument_builder().with_default_value(
          constants.GEOMETRIC_DECAY
      )
      organic_rf_adstock_function = organic_rf_channel_builder(
          **channel_function_map
      )
    else:
      organic_rf_adstock_function = constants.GEOMETRIC_DECAY

    return adstock_hill.AdstockDecaySpec(
        media=media_adstock_function,
        rf=rf_adstock_function,
        organic_media=organic_media_adstock_function,
        organic_rf=organic_rf_adstock_function,
    )

  def populate_cached_properties(self):
    """Eagerly activates all cached properties.

    This is useful for creating a `tf.function` computation graph with this
    Meridian object as part of a captured closure. Within the computation graph,
    internal state mutations are problematic, and so this method freezes the
    object's states before the computation graph is created.
    """
    cls = self.__class__
    # "Freeze" all @cached_property attributes by simply accessing them (with
    # `getattr()`).
    cached_properties = [
        attr
        for attr in dir(self)
        if isinstance(getattr(cls, attr, cls), functools.cached_property)
    ]
    for attr in cached_properties:
      _ = getattr(self, attr)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates an `Mmm` (Marketing Mix Model) proto for Meridian UI.

The MMM proto schema contains parts collected from the core model as well as
analysis results from trained model processors.
"""

import abc
from collections.abc import Collection, Sequence
import dataclasses
import datetime
from typing import TypeVar
import uuid
import warnings

from meridian.analysis import optimizer
from meridian.data import time_coordinates as tc
from meridian.model import model
from mmm.v1 import mmm_pb2 as mmm_pb
from scenarioplanner.converters.dataframe import constants as converter_constants
from schema import mmm_proto_generator
from schema.processors import budget_optimization_processor as bop
from schema.processors import marketing_processor
from schema.processors import model_fit_processor
from schema.processors import model_processor
from schema.processors import reach_frequency_optimization_processor as rfop
from schema.utils import date_range_bucketing


__all__ = [
    "MmmUiProtoGenerator",
    "create_mmm_ui_data_proto",
    "create_tag",
]


_ALLOWED_SPEC_TYPES_FOR_UI = frozenset({
    model_fit_processor.ModelFitSpec,
    marketing_processor.MarketingAnalysisSpec,
    bop.BudgetOptimizationSpec,
})

_SPEC_TYPES_CREATE_SUBSPECS = frozenset({
    marketing_processor.MarketingAnalysisSpec,
    bop.BudgetOptimizationSpec,
    rfop.ReachFrequencyOptimizationSpec,
})

_DATE_RANGE_GENERATORS = frozenset({
    date_range_bucketing.MonthlyDateRangeGenerator,
    date_range_bucketing.QuarterlyDateRangeGenerator,
    date_range_bucketing.YearlyDateRangeGenerator,
})

SpecType = TypeVar("SpecType", bound=model_processor.Spec)
DatedSpecType = TypeVar("DatedSpecType", bound=model_processor.DatedSpec)
OptimizationSpecType = TypeVar(
    "OptimizationSpecType", bound=model_processor.OptimizationSpec
)

_DERIVED_RF_OPT_NAME_PREFIX = "derived RF optimization from "
_DERIVED_RF_OPT_GRID_NAME_PREFIX = "derived_from_"


class MmmUiProtoGenerator:
  """Creates `Mmm` proto for the Meridian Scenario Planner UI (Looker Studio).

  Currently, it only accepts specs for Model Fit, Marketing Analysis, and Budget
  Optimization, but not stand-alone Reach Frequency Optimization specs.
  Reach Frequency Optimization spec will be derived from the Budget Optimization
  spec; this is done so that we can structurally pair them.

  Attributes:
    mmm: A trained Meridian model. A trained model has its posterior
      distributions already sampled.
    specs: A sequence of specs that specify the analyses to run on the model.
    model_id: An optional model identifier.
    time_breakdown_generators: A list of generators that break down the given
      specs by automatically generated time buckets. Currently, this time period
      breakdown is only done on Marketing Analysis specs and Budget Optimization
      specs. All other specs are processed in their original forms. The set of
      default bucketers break down sub-specs with the following time periods:
      [All (original spec's time period), Yearly, Quarterly, Monthly]
  """

  def __init__(
      self,
      mmm: model.Meridian,
      specs: Sequence[SpecType],
      model_id: str = "",
      time_breakdown_generators: Collection[
          type[date_range_bucketing.DateRangeBucketer]
      ] = _DATE_RANGE_GENERATORS,
  ):
    self._mmm = mmm
    self._input_specs = specs
    self._model_id = model_id
    self._time_breakdown_generators = time_breakdown_generators

  @property
  def _time_coordinates(self) -> tc.TimeCoordinates:
    return self._mmm.input_data.time_coordinates

  def __call__(self) -> mmm_pb.Mmm:
    """Creates `Mmm` proto for the Meridian Scenario Planner UI (Looker Studio).

    Returns:
      A proto containing the model kernel at rest and its analysis results given
      user specs.
    """
    seen_group_ids = set()

    copy_specs = []
    for spec in self._input_specs:
      if not any(isinstance(spec, t) for t in _ALLOWED_SPEC_TYPES_FOR_UI):
        raise ValueError(f"Unsupported spec type: {spec.__class__.__name__}")

      if isinstance(spec, bop.BudgetOptimizationSpec):
        group_id = spec.group_id
        if not group_id:
          group_id = str(uuid.uuid4())
          copy_specs.append(dataclasses.replace(spec, group_id=group_id))
        else:
          if group_id in seen_group_ids:
            raise ValueError(
                f"Duplicate group ID found: {group_id}. Please provide a unique"
                " group ID for each Budget Optimization spec."
            )
          seen_group_ids.add(group_id)
          copy_specs.append(spec)

        # If there are RF channels, derive a RF optimization spec from the
        # Budget Optimization spec.
        if self._mmm.input_data.rf_channel is not None:
          copy_specs.append(
              self._derive_rf_opt_spec_from_budget_opt_spec(copy_specs[-1])
          )
      else:
        copy_specs.append(spec)

    sub_specs = []
    for spec in copy_specs:
      to_create_subspecs = self._time_breakdown_generators and any(
          isinstance(spec, t) for t in _SPEC_TYPES_CREATE_SUBSPECS
      )

      if to_create_subspecs:
        dates = self._enumerate_dates_open_end(spec)
        sub_specs.extend(
            _create_subspecs(spec, dates, self._time_breakdown_generators)
        )
      else:
        sub_specs.append(spec)

    return mmm_proto_generator.create_mmm_proto(
        self._mmm,
        sub_specs,
        model_id=self._model_id,
    )

  def _derive_rf_opt_spec_from_budget_opt_spec(
      self,
      budget_opt_spec: bop.BudgetOptimizationSpec,
  ) -> rfop.ReachFrequencyOptimizationSpec:
    """Derives a ReachFrequencyOptimizationSpec from a BudgetOptimizationSpec."""
    rf_opt_name = (
        f"{_DERIVED_RF_OPT_NAME_PREFIX}{budget_opt_spec.optimization_name}"
    )
    rf_opt_grid_name = (
        f"{_DERIVED_RF_OPT_GRID_NAME_PREFIX}{budget_opt_spec.optimization_name}"
    )

    return rfop.ReachFrequencyOptimizationSpec(
        start_date=budget_opt_spec.start_date,
        end_date=budget_opt_spec.end_date,
        date_interval_tag=budget_opt_spec.date_interval_tag,
        optimization_name=rf_opt_name,
        grid_name=rf_opt_grid_name,
        group_id=budget_opt_spec.group_id,
        confidence_level=budget_opt_spec.confidence_level,
    )

  def _enumerate_dates_open_end(
      self, spec: DatedSpecType
  ) -> list[datetime.date]:
    """Enumerates date points with an open end date.

    The date points are enumerated from the data's time coordinates based on the
    spec's start and end dates. The last date point is the exclusive end date as
    same as the spec's end date, if specified.

    Args:
      spec: A dated spec.

    Returns:
      A list of date points.
    """
    inclusive_date_strs = spec.resolver(
        self._mmm.input_data.time_coordinates
    ).resolve_to_enumerated_selected_times()

    if inclusive_date_strs is None:
      dates = self._time_coordinates.all_dates
    else:
      dates = [tc.normalize_date(date_str) for date_str in inclusive_date_strs]

    # If the end date is not specified, compute the exclusive end date based on
    # the last date in the time coordinates.
    exclusive_end_date = spec.end_date or dates[-1] + datetime.timedelta(
        days=self._time_coordinates.interval_days
    )

    dates.append(exclusive_end_date)

    return dates


def create_mmm_ui_data_proto(
    mmm: model.Meridian,
    specs: Sequence[SpecType],
    model_id: str = "",
    time_breakdown_generators: Collection[
        type[date_range_bucketing.DateRangeBucketer]
    ] = _DATE_RANGE_GENERATORS,
) -> mmm_pb.Mmm:
  """Creates `Mmm` proto for the Meridian Scenario Planner UI (Looker Studio).

  Currently, it only accepts specs for Model Fit, Marketing Analysis, and Budget
  Optimization, but not stand-alone Reach Frequency Optimization specs.
  Reach Frequency Optimization spec will be derived from the Budget Optimization
  spec; this is done so that we can structurally pair them.

  Args:
    mmm: A trained Meridian model. A trained model has its posterior
      distributions already sampled.
    specs: A sequence of specs that specify the analyses to run on the model.
    model_id: An optional model identifier.
    time_breakdown_generators: A list of generators that break down the given
      specs by automatically generated time buckets. Currently, this time period
      breakdown is only done on Marketing Analysis specs and Budget Optimization
      specs. All other specs are processed in their original forms. The set of
      default bucketers break down sub-specs with the following time periods:
      [All (original spec's time period), Yearly, Quarterly, Monthly]

  Returns:
    A proto containing the model kernel at rest and its analysis results given
    user specs.
  """
  return MmmUiProtoGenerator(
      mmm,
      specs,
      model_id,
      time_breakdown_generators,
  )()


def create_tag(
    generator_class: type[abc.ABC], start_date: datetime.date
) -> str:
  """Creates a human-readable tag for a spec."""
  if generator_class == date_range_bucketing.YearlyDateRangeGenerator:
    return f"Y{start_date.year}"
  elif generator_class == date_range_bucketing.QuarterlyDateRangeGenerator:
    return f"Y{start_date.year} Q{(start_date.month - 1) // 3 + 1}"
  elif generator_class == date_range_bucketing.MonthlyDateRangeGenerator:
    return f"Y{start_date.year} {start_date.strftime('%b')}"
  else:
    raise ValueError(f"Unsupported generator class: {generator_class}")


def _normalize_optimization_spec_time_info(
    spec: OptimizationSpecType,
    date_interval_tag: str,
) -> OptimizationSpecType:
  """Adds time info to an optimization spec."""
  formatted_date_interval_tag = date_interval_tag.replace(r" ", "_")
  return dataclasses.replace(
      spec,
      group_id=f"{spec.group_id}:{date_interval_tag}",
      optimization_name=f"{spec.optimization_name} for {date_interval_tag}",
      grid_name=f"{spec.grid_name}_{formatted_date_interval_tag}",
  )


def _create_subspecs(
    spec: DatedSpecType,
    date_range: list[datetime.date],
    time_breakdown_generators: Collection[
        type[date_range_bucketing.DateRangeBucketer]
    ],
) -> list[DatedSpecType]:
  """Breaks down a spec into sub-specs for each time bucket."""
  specs = []

  all_period_spec = dataclasses.replace(
      spec,
      date_interval_tag=converter_constants.ANALYSIS_TAG_ALL,
  )
  if isinstance(all_period_spec, model_processor.OptimizationSpec):
    all_period_spec = _normalize_optimization_spec_time_info(
        all_period_spec, converter_constants.ANALYSIS_TAG_ALL
    )
  specs.append(all_period_spec)

  for generator_class in time_breakdown_generators:
    generator = generator_class(date_range)  # pytype: disable=not-instantiable
    date_intervals = generator.generate_date_intervals()
    for start_date, end_date in date_intervals:
      date_interval_tag = create_tag(generator_class, start_date)
      new_spec = dataclasses.replace(
          spec,
          start_date=start_date,
          end_date=end_date,
          date_interval_tag=date_interval_tag,
      )

      if isinstance(new_spec, model_processor.OptimizationSpec):
        new_spec = _normalize_optimization_spec_time_info(
            new_spec, date_interval_tag
        )

        if (
            isinstance(new_spec, bop.BudgetOptimizationSpec)
            and isinstance(new_spec.scenario, optimizer.FixedBudgetScenario)
            and new_spec.scenario.total_budget is not None
        ):
          # TODO: The budget amount should be adjusted based on the
          # budget specified in the `all_period_spec` and the historical spend
          # at the time period.
          new_spec = dataclasses.replace(
              new_spec,
              scenario=optimizer.FixedBudgetScenario(total_budget=None),
          )
          warnings.warn(
              "Using historical spend for budget optimization spec at the"
              f" period of {date_interval_tag}",
          )

      specs.append(new_spec)

  return specs
